## Filename    : CLUnfolder.py
## Author(s)   : Michel Le Borgne
## Created     : 22 mars 2012
## Revision    :
## Source      :
##
## Copyright 2012 : IRISA/IRSET
##
## This library is free software; you can redistribute it and/or modify it
## under the terms of the GNU General Public License as published
## by the Free Software Foundation; either version 2.1 of the License, or
## any later version.
##
## This library is distributed in the hope that it will be useful, but
## WITHOUT ANY WARRANTY, WITHOUT EVEN THE IMPLIED WARRANTY OF
## MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  The software and
## documentation provided here under is on an "as is" basis, and IRISA has
## no obligations to provide maintenance, support, updates, enhancements
## or modifications.
## In no event shall IRISA be liable to any party for direct, indirect,
## special, incidental or consequential damages, including lost profits,
## arising out of the use of this software and its documentation, even if
## IRISA have been advised of the possibility of such damage.  See
## the GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this library; if not, write to the Free Software Foundation,
## Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
##
## The original code contained here was initially developed by:
##
##     Michel Le Borgne.
##     IRISA
##     Symbiose team
##     IRISA  Campus de Beaulieu
##     35042 RENNES Cedex, FRANCE
##
##
## Contributor(s): Geoffroy Andrieux IRSET
##
"""
Main engine for constraint unfolding and solving
"""
#from pyCryptoMS import CryptoMS
from pycryptosat import Solver as CryptoMS
from cadbiom.models.biosignal.translators.gt_visitors import compile_cond, compile_event
from cadbiom.models.clause_constraints.CLDynSys import Clause, Literal
from cadbiom.models.clause_constraints.mcl.MCLSolutions import RawSolution, MCLException
from cadbiom import commons as cm

# C++ API
from _cadbiom import shift_clause, shift_dynamic, forward_code, \
                     forward_init_dynamic

# Standard imports
from logging import DEBUG
import itertools as it

LOGGER = cm.logger()

class CLUnfolder(object):
    """
    Dynamic constraints unfolding management
    ========================================
    Each variable is coded as an integer (in var_code_table and var_list).
    Each clause is represented as a list of signed integers (DIMACS coding).
    During unfolding, clauses are shifted either to the future (forward) or
    to the past (backward). The shift operator is a simple addition
    of self.__shift_step. The shift direction depends only on initializations.
    self.__shift_step depends on the number of variables so it is impossible
    to add variables while generating a trajectory unfolding.

    Glossary
    ========
    - ground variables/ground dimacs code: variables at time 0/their encoding.
    - solution: a solution of the logical dynamic constraint from SAT solver
    - state vector: list of DIMACS code of original (not shifted) variables
      corresponding to a step.
    - trajectory: list of state_vectors
    """

    def __init__(self, sys):
        self.__dyn_sys = sys # symbolic clause dynamic system

        # shift_step equal to the total number of coded variables
        self.__shift_step_init = sys.get_var_number() # shift step of system
        self.__shift_step = self.__shift_step_init    #  current shift/step
        self.__locked = False
        self.__current_step = 1
        self.__steps_before_check = 0
        self.__include_aux_clauses = True

        # assign a DIMACS code number to each variable (invariant)
        self.__lit_cpt = self.__shift_step +1 # counter for aux. var. coding
        self.__var_code_table = dict()            # name -> DIMACS code
        self.__var_list = ['##']                  # DIMACS code -> name - != 0
        cpt = 1
        for svar in sys.base_var_set:
            self.__var_list.append(svar)
            self.__var_code_table[svar] = cpt
            cpt += 1
        assert(cpt == self.__shift_step + 1)

        # same tools for auxiliary variables from property compilation
        self.__aux_code_table = dict()            # name 2 code
        self.__aux_list = []                      # code to name

        # list  no_frontier place DIMACS codes
        self.__no_frontier_init = []
        for nfp in self.__dyn_sys.no_frontier:
            self.__no_frontier_init.append([- self.__var_code_table[nfp]])

        # ordered list of frontier place DIMACS codes for extraction (inv)
        self.__frontier = []
        for frp in self.__dyn_sys.frontier:
            self.__frontier.append(self.__var_code_table[frp])
        self.__frontier.sort()

        # ordered list of input places DIMACS code for extraction (inv)
        self.__inputs = []
        for inp in self.__dyn_sys.inputs:
            self.__inputs.append(self.__var_code_table[inp])
        self.__inputs.sort()

        # ordered list of free clocks DIMACS code for extraction (invariant)
        self.__free_clocks = []
        for fcl in self.__dyn_sys.free_clocks:
            self.__free_clocks.append(self.__var_code_table[fcl])
        self.__free_clocks.sort()

        # properties to be checked
        self.__initial_property = None            # string
        self.__dim_initial = None                 # list of dimacs clauses
        self.__final_property = None              # idem
        self.__dim_final = None
        self.__invariant_property = None          # idem
        self.__dim_invariant = None
        self.__variant_property = None            # list<logic formulas>
        self.__dim_variant = None                 # list<list<dimacs clauses>>
        # list of variant temporal constraints in Dimacs ground code
        self.__list_variant_constraints = None    # list<list<dimacs clauses>>

        # constraints - logical constraints
        # result from unfolding of base constraints
        self.__dynamic_constraints = []        # DIMACS clauses: X' = f(X,H,I)
        self.__initial_constraints = []        # DIMACS clauses: C(X_0)
        self.__final_constraints   = []        # idem: C(X_n)
        self.__invariant_constraints = []      # DIMACS clauses: C(X_i))
        self.__variant_constraints = []        # variant constraints along traj.
        self.__shift_direction = None          # forward or backward

        # statistics on solver
        self.__stats = False
        self.__nb_vars = 0
        self.__nb_clauses = 0



    def reset(self):
        """
        Reset the unfolder before a new query
        """
        self.__initial_property = None
        self.__dim_initial = None
        self.__final_property = None
        self.__dim_final = None
        self.__invariant_property = None
        self.__dim_invariant = None
        self.__variant_property = None
        self.__dim_variant = None
        self.list_variant_constraints = None
        self.__steps_before_check = 0
        self.__shift_direction = None
        self.__locked = False

    def set_stats(self):
        """
        Enable solver statistics
        """
        self.__stats = True
        self.__nb_vars = 0
        self.__nb_clauses = 0

    def unset_stats(self):
        """
        disable solver statistics
        """
        self.__stats = False

    def _stats(self):
        """
        print solver statistics
        """
        print "\n NB Variables in solver: ", self.__nb_vars
        print "NB Clauses in solver: ", self.__nb_clauses

    def set_minimal_steps(self, nb_steps):
        """
        For reachability optimisation. number of shift before checking
        """
        self.__steps_before_check = nb_steps

    def set_include_aux_clauses(self, val):
        """
        Include auxiliary clause to eliminate undesirable solutions
        If A->B has clock h an aux constraint added is h included in h when A
        Must be set to False for inhibitors computations
        """
        self.__include_aux_clauses = val

    # internal variable access for tests
    def get_dynamic_constraints(self):
        """
        For tests: returns coded dynamic constraints
        """
        return self.__dynamic_constraints

    def get_initial_constraints(self):
        """
        For tests: returns coded initial constraints
        """
        return self.__initial_constraints

    def get_invariant_constraints(self):
        """
        For tests: returns coded invariant constraints
        """
        return self.__invariant_constraints

    def get_variant_constraints(self):
        """
        For tests: returns coded variant constraints
        """
        return self.__variant_constraints

    def get_final_constraints(self):
        """
        For tests: returns coded final constraints
        """
        self.__final_constraints

    # variables management
    def var_dimacs_code(self, var_name):
        """
        returns dimacs code of (string) var_name variable
        """
        return self.__var_code_table[var_name]

    def get_system_var_number(self):
        """
        number of variables in the clause constraint dynamical system
        """
        return self.__dyn_sys.get_var_number()

    def get_var_number(self):
        """
        number of principal variables (properties excluded) in the unfolder
        """
        return len(self.__var_list) -1

    def get_var_name(self, var_num):
        """
        @param var_num: DIMACS literal coding of an initial variable
        @return: name of the variable
        """
        if var_num < 0:
            varnum1 = -var_num
        else: varnum1 = var_num
        var_code = (varnum1 - 1) % self.__shift_step + 1
        if var_code <= self.__shift_step_init: # variable from initial system
            return self.__var_list[var_code]
        else: # auxiliary variable introduced by properties compilation
            return self.__aux_list[var_code - self.__shift_step_init -1]
            #raise MCLException("Not a DIMACS code of an initial variable")

    def get_var_indexed_name(self, var_num):
        """
        @param var_num: DIMACS literal coding
        @return: name of the variable with time index
        """
        if var_num < 0:
            varnum1 = -var_num
        else: varnum1 = var_num
        var_code = (varnum1 - 1) % self.__shift_step + 1
        var_step = (varnum1 - var_code)/self.__shift_step
        if var_code <= self.__shift_step_init:
            return self.__var_list[var_code] + '_%s'% var_step
        else:
            index = var_code - self.__shift_step_init -1
            return self.__aux_list[index] + '_%s'% var_step

    def get_frontier(self):
        """
        @return: the DIMACS code of the frontier variables
        """
        return self.__frontier

    def get_free_clocks(self):
        """
        @return: the DIMACS code of the free_clocks variables
        """
        return self.__free_clocks

    def get_inputs(self):
        """
        @return: the DIMACS code of the input variables
        """
        return self.__inputs

    def get_shift_direction(self):
        """
        @return: string "FORWARD" or "BACKWARD"
        """
        return self.__shift_direction

    def get_shift_step(self):
        """
        @return: the shift step ss (if n is X_0 code, n+ss is X_1 code)
        """
        return self.__shift_step

    def get_shift_step_init(self):
        """
        @return: the shift step ss (if n is X_0 code, n+ss is X_1 code)
        """
        return self.__shift_step_init

    def get_current_step(self):
        """
        @return: the current number of unfolding
        """
        return self.__current_step



    # translation from names to num codes
    # the translation depends on the shift direction
    def __forward_code(self, clause):
        """
        numerically code a clause with the numeric code found in self.__var_code_table for
        a base variable x and numeric_code + shift_step for x'
        variable integer coding increases in futur steps
        @param clause: a Clause object
        @return: the DIMACS coding of the forward shifted clause
        """

        # Old API
        # num_clause = []
        # for lit in clause.list_of_literals:
        #     if lit.name[-1] == '`': # t+1 variable
        #         num_clause.append(
        #             -(self.__var_code_table[lit.name[:-1]] + self.__shift_step) \
        #             if not lit.sign \
        #             else (self.__var_code_table[lit.name[:-1]] + self.__shift_step)
        #         )
        #     else: # t variable
        #         num_clause.append(
        #             -self.__var_code_table[lit.name] \
        #             if not lit.sign else self.__var_code_table[lit.name]
        #         )
        # return num_clause

        # New API via C++ module
        #print(forward_code(clause, self.__var_code_table, self.__shift_step))
        return forward_code(clause, self.__var_code_table, self.__shift_step)


    def __backward_code(self, clause):
        """
        numerically code a clause with the numeric code found in self.__var_code_table+ shift_step for
        a base variable x and numeric_code for x'
        variable integer coding increases in past steps
        @param clause: a Clause object
        @return: the DIMACS coding of the backward shifted clause
        """
        num_clause = []
        for lit in clause.list_of_literals:
            if not lit.name[-1] == '`': # t variable
                num_clause.append(-(self.__var_code_table[lit.name] + self.__shift_step) \
                if not lit.sign \
                else (self.__var_code_table[lit.name] + self.__shift_step))
            else:  # t+1 variable
                num_clause.append(-self.__var_code_table[lit.name[:-1]] \
                if not lit.sign \
                else self.__var_code_table[lit.name[:-1]])
        return num_clause

    def __code_clause(self, clause):
        """
        numerically code a clause with the numeric code found in
        self.__var_code_table for variables in the dynamic system and
        self.__aux_code_table for other variables assume all variables
        are basic variables (t=0)
        @warning: MODIFIES SHIFT_STEP if auxiliary variables are present (numeric code)
        """
        num_clause = []
        for lit in clause.list_of_literals:
            name = lit.name
            if self.__var_code_table.has_key(name):
                var_cod = self.__var_code_table[name]
            elif self.__aux_code_table.has_key(name):
                var_cod = self.__aux_code_table[name]
            else: # create an auxiliary variable - modifies shift_step
                self.__shift_step += 1
                self.__aux_code_table[name] = self.__shift_step
                var_cod = self.__shift_step
                self.__aux_list.append(name)
            if lit.sign:
                lit_code = var_cod
            else:
                lit_code = - var_cod
            num_clause.append(lit_code)
        return num_clause

    # dynamics initialisations
    def __forward_init_dynamic(self):
        """
        set dynamic constraints for a forward one step: X1 = f(X0)
        """

        # Old API
        # self.__dynamic_constraints = []

        # num_clause_list = \
        #     [self.__forward_code(clause)
        #      for clause in self.__dyn_sys.list_clauses]
        #
        # if self.__include_aux_clauses:
        #     num_clause_list += \
        #         [self.__forward_code(clause)
        #          for clause in self.__dyn_sys.aux_list_clauses]
        #
        # self.__dynamic_constraints.append(num_clause_list)

        # New API via C++ module
        # TODO: take a look at __backward_init_dynamic & __backward_code
        if self.__include_aux_clauses:
            self.__dynamic_constraints = \
                forward_init_dynamic(self.__dyn_sys.list_clauses,
                                     self.__var_code_table,
                                     self.__shift_step,
                                     self.__dyn_sys.aux_list_clauses)
        else:
            self.__dynamic_constraints = \
                forward_init_dynamic(self.__dyn_sys.list_clauses,
                                     self.__var_code_table,
                                     self.__shift_step)

    def __backward_init_dynamic(self):
        """
        set dynamic constraints for a forward one step: X0 = f(X1)
        """
        self.__dynamic_constraints = []

        num_clause_list = \
            [self.__backward_code(clause)
             for clause in self.__dyn_sys.list_clauses]

        if self.__include_aux_clauses:
            num_clause_list += \
                [self.__backward_code(clause)
                 for clause in self.__dyn_sys.aux_list_clauses]

        self.__dynamic_constraints.append(num_clause_list)

    # shifting: implementation of the shift operator
    def __shift_clause(self, ncl):
        """
        shift a clause one step
        @param ncl: DIMACS clause
        @warning: lock the unfolder
        """
        self.__locked = True  # shift_step must be frozen

        # Old API
        # Less efficient with abs()
        # return [(abs(lit) + self.__shift_step) * (-1 if lit < 0 else 1) for lit in ncl]
        # More efficient with ternary assignment
        # return [(lit + self.__shift_step) if lit > 0 else (lit - self.__shift_step)
        #         for lit in ncl]

        # New API via C++ module
        return shift_clause(ncl, self.__shift_step)

    def __m_shift_clause(self, ncl, nb_steps):
        """
        shift a clause many step
        @param ncl: DIMACS clause
        @param nb_steps: number of shifts
        @warning: lock the unfolder
        """
        self.__locked = True  # shift_step must be frozen

        return [(lit + self.__shift_step * nb_steps) if lit > 0
                else (lit - self.__shift_step * nb_steps) for lit in ncl]

    def __shift_dynamic(self):
        """
        Shift clauses representing the dynamics X' = f(X,I,C)
        """

        # Old API
        # self.__dynamic_constraints.append(
        #   [self.__shift_clause(clause)
        #    for clause in self.__dynamic_constraints[-1]]
        # )

        # New API via C++ module
        self.__dynamic_constraints.append(
            shift_dynamic(
                self.__dynamic_constraints[-1],
                self.__shift_step
            )
        )

    def __shift_initial(self):
        """
        Shift initialisation condition
        """

        self.__initial_constraints = \
            [self.__shift_clause(clause)
             for clause in self.__initial_constraints]

    def __shift_final(self):
        """
        Shift final property
        """

        self.__final_constraints = \
            [self.__shift_clause(clause)
             for clause in self.__final_constraints]


    def __shift_invariant(self):
        """
        Shift invariant property
        """
        if not self.__invariant_constraints == []:

            # Old API
            # self.__invariant_constraints.append(
            #     [self.__shift_clause(clause)
            #      for clause in self.__invariant_constraints[-1]]
            # )

            # New API via C++ module
            self.__invariant_constraints.append(
                shift_dynamic(
                    self.__invariant_constraints[-1],
                    self.__shift_step
                )
            )

    def __shift_variant(self):
        """
        shift variant property - depends on unfolding direction
        """
        if not self.__list_variant_constraints:
            return
        if self.__shift_direction == "FORWARD":
            index = self.__current_step # constraint t0 already included
            current_constraint = self.__list_variant_constraints[index]
            # shift constraint at current step and add to var_constraints
            for clause in current_constraint:
                s_clause = self.__m_shift_clause(clause, index)
                self.__variant_constraints.append(s_clause)
                return
        elif self.__shift_direction == 'BACKWARD':
            raise MCLException("Not yet implemented")
        else:
            raise MCLException("Shift incoherent data: "+self.__shift_direction)


    def shift(self):
        """
        assume shift direction is set - shift one step
        """
        self.__shift_dynamic()
        self.__shift_invariant()
        self.__shift_variant()
        if self.__shift_direction == 'FORWARD':
            self.__shift_final()
        elif self.__shift_direction == 'BACKWARD':
            self.__shift_initial()
        else:
            raise MCLException("Shift incoherent data: "+self.__shift_direction)
        self.__current_step += 1

    # coding of properties
    def __compile_property(self, property_text):
        """
        compile a property (logical formula) into clauses
        type checking uses the symbol table of dyn_sys
        error reporting through the dyn_sys reporter
        MODIFIES __lit_cpt for numbering of auxiliary variables (not coding)
        """
        if self.__locked:
            raise MCLException("Trying to compile property while unfolder is locked")

        # syntax analyser and type checker
        # property_text, symb_t, reporter
        tree_prop = compile_cond(property_text,
                                 self.__dyn_sys.symb_tab,
                                 self.__dyn_sys.report)
        # avoid name collisions of aux var
        prop_visitor = CLPropertyVisitor(self.__lit_cpt)
        tree_prop.accept(prop_visitor)
        self.__lit_cpt = prop_visitor.cpt # avoid name collisions of aux var
        return prop_visitor.clauses

        # coding of properties

    def __compile_event(self, property_text):
        """
        compile an event (biosignal expression) into clauses
        type checking uses the symbol table of dyn_sys
        error reporting through the dyn_sys reporter
        MODIFIES __lit_cpt for numbering of auxiliary variables (not coding)
        """
        if self.__locked:
            mess = "Trying to compile property while unfolder is locked"
            raise MCLException(mess)
        reporter = self.__dyn_sys.report
        symb_t = self.__dyn_sys.symb_tab
        # syntax analyser and type checker
        (tree_prop, ste, fcl) = compile_event(property_text, symb_t,
                                              True,  reporter)
        # avoid name collisions of aux var
        prop_visitor = CLPropertyVisitor(self.__lit_cpt)
        tree_prop.accept(prop_visitor)
        self.__lit_cpt = prop_visitor.cpt # avoid name collisions of aux var
        return prop_visitor.clauses

    def set_initial_property(self, property_text):
        """
        @param property_text: string - litteral boolean expression
        """
        self.__initial_property = property_text

    def set_dim_initial_property(self, dimacs_clause_list):
        """
        set initial property with a DIMACS representation
        @param dimacs_clause_list:
        """
        self.__dim_initial = dimacs_clause_list

    def __init_initial_constraint_0(self, no_frontier_init=True):
        """
        if initial property is set, generate constraints clauses in numeric form
        base variable form - we have to wait until all variables are num coded
        before shifting anything!!
        """
# OLD
#        self.__initial_constraints = []
#        if no_frontier_init:
#            # initialize no frontier places to False
#            for nfp in self.__no_frontier_init:
#                self.__initial_constraints.append(nfp)
#        if self.__initial_property:
#            # compile initial property
#            l_clauses = self.__compile_property(self.__initial_property)
#            # compile numeric form
#            for clause in l_clauses:
#                self.__initial_constraints.append(self.__code_clause(clause))
#        # dimacs aux initial properties
#        dim_init = self.__dim_initial
#        if dim_init:
#            self.__initial_constraints = self.__initial_constraints + dim_init

        self.__initial_constraints = list()

        # initialize no frontier places to False
        if no_frontier_init:
            self.__initial_constraints += [elem for elem in self.__no_frontier_init]

        if self.__initial_property:
            # compile initial property
            # compile numeric form
            self.__initial_constraints += [self.__code_clause(clause) for clause in self.__compile_property(self.__initial_property)]

        # dimacs aux initial properties
        if self.__dim_initial:
            self.__initial_constraints += self.__dim_initial

    def set_final_property(self, property_text):
        """
        @param property_text: string - litteral boolean expression
        """
        self.__final_property = property_text

    def set_dim_final_property(self, dimacs_clause_list):
        """
        set final property with a DIMACS representation
        @param dimacs_clause_list:
        """
        self.__dim_final = dimacs_clause_list

    def __init_final_constraint_0(self):
        """
        if final property is set, generate constraints clauses in numeric form
        base variable used - we have to wait until all variables are num coded
        before shifting anything!!
        """

        self.__final_constraints = []
        if self.__final_property:
            # compile initial (X0) property
            l_clauses = self.__compile_property(self.__final_property)
            # compile numeric form
            for clause in l_clauses:
                self.__final_constraints.append(self.__code_clause(clause))
        dim_fin = self.__dim_final
        if dim_fin:
            self.__final_constraints = self.__final_constraints + dim_fin

    def set_current_property(self, property_text):
        """
        @param property_text: string - litteral boolean expression
        """

        self.__invariant_property = property_text

    def set_dim_current_property(self, dimacs_clause_list):
        """
        set final property with a DIMACS representation
        @param dimacs_clause_list:
        """

        self.__dim_invariant = dimacs_clause_list

    def __init_invariant_constraint_0(self):
        """
        if trajectory property is set, generate constraints clauses in numeric form
        base variable used - we have to wait until all variables are num coded
        before shifting anything!!
        """
        self.__invariant_constraints = []
        if self.__invariant_property:
            # compile initial (X0) property
            l_clauses = self.__compile_property(self.__invariant_property)
            # compile numeric form
            init_const = []
            for clause in l_clauses:
                init_const.append(self.__code_clause(clause))
            self.__invariant_constraints.append(init_const)
        d_inv = self.__dim_invariant
        if d_inv:
            self.__invariant_constraints = self.__invariant_constraints + d_inv

    def set_variant_property(self, var_prop):
        """
        @param var_prop: a list of logical formulas representing time varying
        constraints. Each constraint uses ground (not shifted) variables.
        The order of the list of constraints is always time increasing.
        """
        self.__variant_property = var_prop

    def set_dim_variant_property(self, dim_var_prop):
        """
        @param dim_var_prop: a list list of dimacs clauses representing time
        varying constraints. Each constraint uses ground (not shifted)
        variable codes.
        """
        self.__dim_variant = dim_var_prop

    def __init_variant_constraints_0(self):
        """
        if variant_property is set, compile each property into clauses and
        encode these clauses. Clauses already in dimacs form are added.
        The two variant constraint forms must be compatible (same length)
        """
        # coding of variant properties
        if self.__variant_property:
            self.__list_variant_constraints = []
            for prop in self.__variant_property:
                # compile initial (X0) property
                l_clauses = self.__compile_event(prop)
                # encode the clauses
                dim_cl = []
                for clause in l_clauses:
                    dim_cl.append(self.__code_clause(clause))
                self.__list_variant_constraints.append(dim_cl)

            # add variant properties in dimacs form
            if self.__dim_variant:
                vp_length = len(self.__variant_property)
                if vp_length != len(self.__dim_variant):
                    mess = "Incoherent variant properties"
                    raise MCLException(mess)

                for i in range(vp_length):
                    c_var = self.__list_variant_constraints[i]
                    c_var = c_var + self.__dim_variant[i]
                    self.__list_variant_constraints[i] = c_var
        else:
            if self.__dim_variant:
                self.__list_variant_constraints = self.__dim_variant
        # initialisation
        if self.__list_variant_constraints:
            if self.__shift_direction == "FORWARD":
                self.__variant_constraints = self.__list_variant_constraints[0]
            elif  self.__shift_direction == "BACKWARD":
                raise MCLException("Not yet implemented")
            else:
                raise MCLException("Shift incoherent data: "
                                   + self.__shift_direction)

    # whole initialisations
    def init_forward_unfolding(self):
        """
        initialisation before generating constraints - forward trajectory
        """
        self.__shift_direction = 'FORWARD'
        self.__current_step = 1
        self.__shift_step = self.__shift_step_init  # back to basic!
        self.__aux_code_table = dict()              # flush auxiliary variables
        self.__aux_list = []                        # idem

        # init properties to generate all variable num codes
        self.__init_initial_constraint_0()
        self.__init_final_constraint_0()
        self.__init_invariant_constraint_0()
        self.__init_variant_constraints_0()

        # now shifting is possible
        self.__forward_init_dynamic()
        self.__shift_final()
        self.__shift_invariant()


    def init_backward_unfolding(self):
        """
        initialisation before generating constraints - backward trajectory
        """
        self.__shift_direction = 'BACKWARD'
        self.__current_step = 0
        self.__shift_step = self.__shift_step_init  # back to basic!
        self.__aux_code_table = dict()              # flush auxiliary variables
        self.__aux_list = []                        # idem

        # init properties to generate all variable num codes
        self.__init_initial_constraint_0()
        self.__init_final_constraint_0()
        self.__init_invariant_constraint_0()
        self.__init_variant_constraints_0()

        # now shifting is possible
        self.__backward_init_dynamic()
        self.__shift_initial()
        self.__shift_invariant()

    # for debug
    def vars_in_clause(self, clause):
        """
        DEBUG
        """
        lvar = []
        for var in clause:
            lvar.append(self.get_var_indexed_name(var))
        return lvar


    # solver interface
    def __load_solver(self, solv):
        """
        add all the current clauses in solver solv
        """

        # New API via C++ module
        solv.add_clauses(self.__final_constraints)
        solv.add_clauses(it.chain(*self.__invariant_constraints))
        solv.add_clauses(self.__variant_constraints)
        solv.add_clauses(it.chain(*self.__dynamic_constraints))
        solv.add_clauses(self.__initial_constraints)

        # Less efficient alternative
        # solv.add_clauses([clause for lcl in self.__invariant_constraints
        #                   for clause in lcl])
        # solv.add_clauses([clause for lcl in self.__dynamic_constraints
        #                   for clause in lcl])

        # Old API
        # Clauses are added one by one
        # final
        #[solv.add_clause(clause) for clause in self.__final_constraints]

        # trajectory invariant
        #[solv.add_clause(clause) for lcl in self.__invariant_constraints
        # for clause in lcl]

        # trajectory variant
        #[solv.add_clause(clause) for clause in self.__variant_constraints]

        # dynamics
        #[solv.add_clause(clause) for lcl in self.__dynamic_constraints
        # for clause in lcl]

        # initial
        #[solv.add_clause(clause) for clause in self.__initial_constraints]

        # LOGGING
        if LOGGER.getEffectiveLevel() == DEBUG:
            # final
            LOGGER.debug("Load new solver !!")
            LOGGER.debug(">> final: " + str(len(self.__final_constraints)))
            LOGGER.debug(str(self.__final_constraints))

            # trajectory invariant
            LOGGER.debug(">> trajectory inv: " + str(len(self.__invariant_constraints)))
            LOGGER.debug(str(self.__invariant_constraints))

            # trajectory variant
            LOGGER.debug(">> trajectory var: " + str(len(self.__variant_constraints)))
            LOGGER.debug(str(self.__variant_constraints))

            # dynamics
            LOGGER.debug(">> dynamics: " + str(len(self.__dynamic_constraints)))
            LOGGER.debug(str(self.__dynamic_constraints))

            # initial
            LOGGER.debug(">> initial: " + str(len(self.__initial_constraints)))
            LOGGER.debug(str(self.__initial_constraints))


    def __constraints_satisfied(self):
        """
        @return: boolean
        """
        LOGGER.debug("__constraints_satisfied ?")
        solver = CryptoMS()
        # Load Solver and all clauses
        self.__load_solver(solver)
        # If stats are activated (never besides in test environment):
        # sync local data (nb vars, nb clauses with solver state)
        if self.__stats:
            if solver.nb_vars() > self.__nb_vars:
                self.__nb_vars = solver.nb_vars()
            if solver.nb_clauses() > self.__nb_clauses:
                self.__nb_clauses = solver.nb_clauses()
        # Is problem satisfiable ?
        return solver.is_satisfiable()


    def __msolve_constraints(self, max_sol, vvars):
        """
        @param max_sol: int - the max number of solution returned
        @param vvars: variables for which the solver must find different solutions(dimacs code)
        @return: a list of RawSolution
        """
        solver = CryptoMS()
        self.__load_solver(solver)

        if self.__stats:
            if solver.nb_vars() > self.__nb_vars:
                self.__nb_vars = solver.nb_vars()
            if solver.nb_clauses() > self.__nb_clauses:
                self.__nb_clauses = solver.nb_clauses()

        if LOGGER.getEffectiveLevel() == DEBUG:
            LOGGER.debug("__msolve_constraints :: vvars : " + str(vvars))
            LOGGER.debug("__msolve_constraints :: max_sol : " + str(max_sol))
            lintsol = solver.msolve_selected(max_sol, vvars)
            LOGGER.debug("__msolve_constraints :: lintsol : " + str(lintsol))
            return [RawSolution(solint, self) for solint in lintsol]

        return [RawSolution(solint, self)
                for solint in solver.msolve_selected(max_sol, vvars)]

    # dynamic properties
    def squery_is_satisfied(self, max_step):
        """
        Ask the SAT solver
        @param max_step: int - the horizon
        """
        # initialization
        self.init_forward_unfolding()
        # horizon adjustment
        if self.__list_variant_constraints:
            max_step = min(max_step, len(self.__list_variant_constraints)-1)
        # loop
        if self.__invariant_property and not self.__final_property :
            while self.__current_step <= max_step:
                self.shift()
            return self.__constraints_satisfied()
        else :
            while self.__current_step <= max_step:
                if self.__constraints_satisfied():
                    return True
                self.shift()
            return self.__constraints_satisfied()



    def squery_solve(self, vvars, max_step, max_sol):
        """
        Assert a query is loaded (start_condition, inv_condition, final_condition)
        @param vvars: variables on which different solutions must differ
        @param max_step: bounded horizon for computations
        @param max_sol: max number of solutions to be computed

        @return: list of RawSolution objects
        """
        if LOGGER.getEffectiveLevel() == DEBUG:
            LOGGER.debug("squery_solve :: vvars : " + str(vvars))
        # initialization
        self.init_forward_unfolding()
        # horizon adjustment
        if self.__list_variant_constraints:
            max_step = min(max_step, len(self.__list_variant_constraints)-1)
        # loop
        if self.__invariant_property and not self.__final_property :
            while self.__current_step <= max_step:
                self.shift()
            # return l_rawsol
            return self.__msolve_constraints(max_sol, vvars)

        else : # reachability
            l_rawsol = []
            # check and correct for incoherent data on step number
            if self.__steps_before_check > max_step:
                self.__steps_before_check = max_step - 1
            # shift without checking (optimization)
            while self.__current_step <= self.__steps_before_check:
                self.shift()
            # search for solutions at each step
            while self.__current_step <= max_step:
                lsol = self.__msolve_constraints(max_sol, vvars)
                if len(lsol)>0:
                    l_rawsol = l_rawsol + lsol
                self.shift()
            return l_rawsol


###############################################################################
class CLPropertyVisitor(object):
    """
    SigExpression on states and events into a set of clauses
    Type checking must be done.
    """

    def __init__(self, aux_cpt=0):
        self.cpt = aux_cpt         # for auxiliary variable naming
        self.clauses = []          # generated clauses
        self.top = True            # root of the formula?

    def visit_sig_ident(self, tnode):
        """
        ident -> literal
        """
        name = tnode.name
        new_lit = Literal(name, True)
        if self.top:
            self.clauses.append(Clause([new_lit]))
            return None
        return new_lit

    def visit_sig_not(self, node):
        """
        translate a not
        """
        top = self.top
        self.top = False
        # compile operand to nl <=> formula
        newl = node.operand.accept(self)
        notnl = newl.lit_not()
        if top: # we are at the root
            self.clauses.append(Clause([notnl]))
            return None
        return notnl # if we are not at the root

    def visit_sig_sync(self, binnode):
        """
        translate a binary (or/and) expression
        """
        top = self.top
        self.top = False
        # the internal nodes will be affected with the name '_lit'
        name = '_lit'+str(self.cpt)
        self.cpt += 1
        newl = Literal(name, True)
        notnl = Literal(name, False)
        # recursive visits
        operator = binnode.operator
        nl1 = binnode.left_h.accept(self)
        notnl1 = nl1.lit_not()
        nl2 = binnode.right_h.accept(self)
        notnl2 = nl2.lit_not()
        # clauses generation
        if operator == 'and':   # x = lh and rh
            self.clauses.append(Clause([nl1, notnl]))           # not x or not lh
            self.clauses.append(Clause([nl2, notnl]))           # not x or not rh
            self.clauses.append(Clause([notnl1, notnl2, newl])) # x or not lh or not rh

        if operator == 'or':    # x = lh or rh
            self.clauses.append(Clause([notnl1, newl]))        # x or not lh
            self.clauses.append(Clause([notnl2, newl]))        # x or not rh
            self.clauses.append(Clause([nl1, nl2, notnl]))     # not x or not lh or not rh

#        self.clauses.append(cl1)
#        self.clauses.append(cl2)
#        self.clauses.append(cl3)
        if top:
            self.clauses.append(Clause([newl]))
            return None
        return newl

    def visit_sig_const(self, exp):
        """
        translate a constant expression
        """
        top = self.top
        if top:
            if exp.is_const_false():
                raise TypeError("No constant False property allowed")
            else: # always present or satisfied
                return None
        # the internal nodes will be affected with the name '_lit'
        name = '_lit'+str(self.cpt)
        self.cpt += 1
        newl = Literal(name, True)

        # clause generation
        if exp.value:
            clause = newl  # x = True
        else:
            clause = newl.lit_not()
        self.clauses.append(Clause([clause]))
        return newl

    # time operators
    def visit_sig_default(self, exp):
        """
        default translation
        """
        top = self.top
        self.top = False
        # the internal nodes will be affected the name '_lit'
        name = '_lit'+str(self.cpt)
        self.cpt += 1
        newl = Literal(name, True)
        notnl = Literal(name, False)
        # recursive visits
        nl1 = exp.left_h.accept(self)
        notnl1 = nl1.lit_not()
        nl2 = exp.right_h.accept(self)
        notnl2 = nl2.lit_not()
        # clause generation: x = lh or rh
        cl1 = Clause([notnl1, newl])        # x or not lh
        cl2 = Clause([notnl2, newl])        # x or not rh
        cl3 = Clause([nl1, nl2, notnl])     # not x or not lh or not rh

        self.clauses.append(cl1)
        self.clauses.append(cl2)
        self.clauses.append(cl3)
        if top:
            self.clauses.append(Clause([newl]))
            return None
        return newl

    def visit_sig_when(self, exp):
        """
        when translation
        """
        top = self.top
        self.top = False
        # the internal nodes will be affected the name '_lit'
        name = '_lit'+str(self.cpt)
        self.cpt += 1
        newl = Literal(name, True)
        notnl = Literal(name, False)
        # recursive visits
        nl1 = exp.left_h.accept(self)
        notnl1 = nl1.lit_not()
        nl2 = exp.right_h.accept(self)
        notnl2 = nl2.lit_not()
        # clause generation: x = lh and rh
        cl1 = Clause([nl1, notnl])           # not x or not lh
        cl2 = Clause([nl2, notnl])           # not x or not rh
        cl3 = Clause([notnl1, notnl2, newl]) # x or not lh or not rh

        self.clauses.append(cl1)
        self.clauses.append(cl2)
        self.clauses.append(cl3)
        if top:
            self.clauses.append(Clause([newl]))
            return None
        return newl




# $ANTLR 3.1.2 /home/gandrieu/cadbiomworkspace/Cadbiom_des_new/models/guard_transitions/condexp.g 2012-09-11 14:55:48

import sys
from cadbiom.antlr3 import *
from cadbiom.antlr3.compat import set, frozenset




# for convenience in actions
HIDDEN = BaseRecognizer.HIDDEN

# token types
F=12
WS=13
LETTER=15
OR=5
DOL=4
T=11
IDENT=8
DIGIT=16
NOT=7
COMMENT=14
PD=10
AND=6
EOF=-1
PG=9

# token names
tokenNames = [
    "<invalid>", "<EOR>", "<DOWN>", "<UP>",
    "DOL", "OR", "AND", "NOT", "IDENT", "PG", "PD", "T", "F", "WS", "COMMENT",
    "LETTER", "DIGIT"
]




class condexpParser(Parser):
    grammarFileName = "/home/gandrieu/cadbiomworkspace/Cadbiom_des_new/models/guard_transitions/condexp.g"
    antlr_version = version_str_to_tuple("3.1.2")
    antlr_version_str = "3.1.2"
    tokenNames = tokenNames

    def __init__(self, input, state=None):
        if state is None:
            state = RecognizerSharedState()

        Parser.__init__(self, input, state)














    # $ANTLR start "sig_bool"
    # /home/gandrieu/cadbiomworkspace/Cadbiom_des_new/models/guard_transitions/condexp.g:13:1: sig_bool returns [idents] : (id1= sig_bool1 DOL | DOL );
    def sig_bool(self, ):

        idents = None

        id1 = None


        idents = set([])
        try:
            try:
                # /home/gandrieu/cadbiomworkspace/Cadbiom_des_new/models/guard_transitions/condexp.g:15:9: (id1= sig_bool1 DOL | DOL )
                alt1 = 2
                LA1_0 = self.input.LA(1)

                if ((NOT <= LA1_0 <= PG) or (T <= LA1_0 <= F)) :
                    alt1 = 1
                elif (LA1_0 == DOL) :
                    alt1 = 2
                else:
                    nvae = NoViableAltException("", 1, 0, self.input)

                    raise nvae

                if alt1 == 1:
                    # /home/gandrieu/cadbiomworkspace/Cadbiom_des_new/models/guard_transitions/condexp.g:15:11: id1= sig_bool1 DOL
                    pass
                    self._state.following.append(self.FOLLOW_sig_bool1_in_sig_bool65)
                    id1 = self.sig_bool1()

                    self._state.following.pop()
                    self.match(self.input, DOL, self.FOLLOW_DOL_in_sig_bool67)
                    #action start
                    idents = id1
                    #action end


                elif alt1 == 2:
                    # /home/gandrieu/cadbiomworkspace/Cadbiom_des_new/models/guard_transitions/condexp.g:16:11: DOL
                    pass
                    self.match(self.input, DOL, self.FOLLOW_DOL_in_sig_bool81)



            except RecognitionException, re:
                self.reportError(re)
                self.recover(self.input, re)
        finally:

            pass

        return idents

    # $ANTLR end "sig_bool"


    # $ANTLR start "sig_bool1"
    # /home/gandrieu/cadbiomworkspace/Cadbiom_des_new/models/guard_transitions/condexp.g:19:1: sig_bool1 returns [idents] : id1= sig_bool_and ( OR id2= sig_bool_and )* ;
    def sig_bool1(self, ):

        idents = None

        id1 = None

        id2 = None


        idents = set([])
        try:
            try:
                # /home/gandrieu/cadbiomworkspace/Cadbiom_des_new/models/guard_transitions/condexp.g:21:9: (id1= sig_bool_and ( OR id2= sig_bool_and )* )
                # /home/gandrieu/cadbiomworkspace/Cadbiom_des_new/models/guard_transitions/condexp.g:21:11: id1= sig_bool_and ( OR id2= sig_bool_and )*
                pass
                self._state.following.append(self.FOLLOW_sig_bool_and_in_sig_bool1146)
                id1 = self.sig_bool_and()

                self._state.following.pop()
                #action start
                idents = id1
                #action end
                # /home/gandrieu/cadbiomworkspace/Cadbiom_des_new/models/guard_transitions/condexp.g:22:11: ( OR id2= sig_bool_and )*
                while True: #loop2
                    alt2 = 2
                    LA2_0 = self.input.LA(1)

                    if (LA2_0 == OR) :
                        alt2 = 1


                    if alt2 == 1:
                        # /home/gandrieu/cadbiomworkspace/Cadbiom_des_new/models/guard_transitions/condexp.g:22:12: OR id2= sig_bool_and
                        pass
                        self.match(self.input, OR, self.FOLLOW_OR_in_sig_bool1161)
                        self._state.following.append(self.FOLLOW_sig_bool_and_in_sig_bool1165)
                        id2 = self.sig_bool_and()

                        self._state.following.pop()
                        #action start
                        idents = idents | id2
                        #action end


                    else:
                        break #loop2






            except RecognitionException, re:
                self.reportError(re)
                self.recover(self.input, re)
        finally:

            pass

        return idents

    # $ANTLR end "sig_bool1"


    # $ANTLR start "sig_bool_and"
    # /home/gandrieu/cadbiomworkspace/Cadbiom_des_new/models/guard_transitions/condexp.g:26:1: sig_bool_and returns [idents] : id1= sig_primary ( AND id2= sig_primary )* ;
    def sig_bool_and(self, ):

        idents = None

        id1 = None

        id2 = None


        idents = set([])
        try:
            try:
                # /home/gandrieu/cadbiomworkspace/Cadbiom_des_new/models/guard_transitions/condexp.g:28:9: (id1= sig_primary ( AND id2= sig_primary )* )
                # /home/gandrieu/cadbiomworkspace/Cadbiom_des_new/models/guard_transitions/condexp.g:28:11: id1= sig_primary ( AND id2= sig_primary )*
                pass
                self._state.following.append(self.FOLLOW_sig_primary_in_sig_bool_and229)
                id1 = self.sig_primary()

                self._state.following.pop()
                #action start
                idents = id1
                #action end
                # /home/gandrieu/cadbiomworkspace/Cadbiom_des_new/models/guard_transitions/condexp.g:29:11: ( AND id2= sig_primary )*
                while True: #loop3
                    alt3 = 2
                    LA3_0 = self.input.LA(1)

                    if (LA3_0 == AND) :
                        alt3 = 1


                    if alt3 == 1:
                        # /home/gandrieu/cadbiomworkspace/Cadbiom_des_new/models/guard_transitions/condexp.g:29:12: AND id2= sig_primary
                        pass
                        self.match(self.input, AND, self.FOLLOW_AND_in_sig_bool_and244)
                        self._state.following.append(self.FOLLOW_sig_primary_in_sig_bool_and248)
                        id2 = self.sig_primary()

                        self._state.following.pop()
                        #action start
                        idents =  idents | id2
                        #action end


                    else:
                        break #loop3






            except RecognitionException, re:
                self.reportError(re)
                self.recover(self.input, re)
        finally:

            pass

        return idents

    # $ANTLR end "sig_bool_and"


    # $ANTLR start "sig_primary"
    # /home/gandrieu/cadbiomworkspace/Cadbiom_des_new/models/guard_transitions/condexp.g:33:1: sig_primary returns [idents] : ( NOT id1= sig_primary | id4= sig_constant | id2= IDENT | PG id3= sig_bool1 PD );
    def sig_primary(self, ):

        idents = None

        id2 = None
        id1 = None

        id4 = None

        id3 = None


        idents = set([])
        try:
            try:
                # /home/gandrieu/cadbiomworkspace/Cadbiom_des_new/models/guard_transitions/condexp.g:35:9: ( NOT id1= sig_primary | id4= sig_constant | id2= IDENT | PG id3= sig_bool1 PD )
                alt4 = 4
                LA4 = self.input.LA(1)
                if LA4 == NOT:
                    alt4 = 1
                elif LA4 == T or LA4 == F:
                    alt4 = 2
                elif LA4 == IDENT:
                    alt4 = 3
                elif LA4 == PG:
                    alt4 = 4
                else:
                    nvae = NoViableAltException("", 4, 0, self.input)

                    raise nvae

                if alt4 == 1:
                    # /home/gandrieu/cadbiomworkspace/Cadbiom_des_new/models/guard_transitions/condexp.g:35:11: NOT id1= sig_primary
                    pass
                    self.match(self.input, NOT, self.FOLLOW_NOT_in_sig_primary310)
                    self._state.following.append(self.FOLLOW_sig_primary_in_sig_primary314)
                    id1 = self.sig_primary()

                    self._state.following.pop()
                    #action start
                    idents = id1
                    #action end


                elif alt4 == 2:
                    # /home/gandrieu/cadbiomworkspace/Cadbiom_des_new/models/guard_transitions/condexp.g:38:11: id4= sig_constant
                    pass
                    self._state.following.append(self.FOLLOW_sig_constant_in_sig_primary356)
                    id4 = self.sig_constant()

                    self._state.following.pop()
                    #action start
                    idents = id4
                    #action end


                elif alt4 == 3:
                    # /home/gandrieu/cadbiomworkspace/Cadbiom_des_new/models/guard_transitions/condexp.g:41:11: id2= IDENT
                    pass
                    id2=self.match(self.input, IDENT, self.FOLLOW_IDENT_in_sig_primary388)
                    #action start
                    idents = set([id2.text.encode("utf8")])
                    #action end


                elif alt4 == 4:
                    # /home/gandrieu/cadbiomworkspace/Cadbiom_des_new/models/guard_transitions/condexp.g:44:11: PG id3= sig_bool1 PD
                    pass
                    self.match(self.input, PG, self.FOLLOW_PG_in_sig_primary429)
                    self._state.following.append(self.FOLLOW_sig_bool1_in_sig_primary434)
                    id3 = self.sig_bool1()

                    self._state.following.pop()
                    self.match(self.input, PD, self.FOLLOW_PD_in_sig_primary436)
                    #action start
                    idents = id3
                    #action end



            except RecognitionException, re:
                self.reportError(re)
                self.recover(self.input, re)
        finally:

            pass

        return idents

    # $ANTLR end "sig_primary"


    # $ANTLR start "sig_constant"
    # /home/gandrieu/cadbiomworkspace/Cadbiom_des_new/models/guard_transitions/condexp.g:48:1: sig_constant returns [idents] : ( T | F );
    def sig_constant(self, ):

        idents = None

        idents = set([])
        try:
            try:
                # /home/gandrieu/cadbiomworkspace/Cadbiom_des_new/models/guard_transitions/condexp.g:50:10: ( T | F )
                alt5 = 2
                LA5_0 = self.input.LA(1)

                if (LA5_0 == T) :
                    alt5 = 1
                elif (LA5_0 == F) :
                    alt5 = 2
                else:
                    nvae = NoViableAltException("", 5, 0, self.input)

                    raise nvae

                if alt5 == 1:
                    # /home/gandrieu/cadbiomworkspace/Cadbiom_des_new/models/guard_transitions/condexp.g:50:12: T
                    pass
                    self.match(self.input, T, self.FOLLOW_T_in_sig_constant491)
                    #action start
                    idents = set([])
                    #action end


                elif alt5 == 2:
                    # /home/gandrieu/cadbiomworkspace/Cadbiom_des_new/models/guard_transitions/condexp.g:51:12: F
                    pass
                    self.match(self.input, F, self.FOLLOW_F_in_sig_constant506)
                    #action start
                    idents = set([])
                    #action end



            except RecognitionException, re:
                self.reportError(re)
                self.recover(self.input, re)
        finally:

            pass

        return idents

    # $ANTLR end "sig_constant"


    # Delegated rules




    FOLLOW_sig_bool1_in_sig_bool65 = frozenset([4])
    FOLLOW_DOL_in_sig_bool67 = frozenset([1])
    FOLLOW_DOL_in_sig_bool81 = frozenset([1])
    FOLLOW_sig_bool_and_in_sig_bool1146 = frozenset([1, 5])
    FOLLOW_OR_in_sig_bool1161 = frozenset([7, 8, 9, 11, 12])
    FOLLOW_sig_bool_and_in_sig_bool1165 = frozenset([1, 5])
    FOLLOW_sig_primary_in_sig_bool_and229 = frozenset([1, 6])
    FOLLOW_AND_in_sig_bool_and244 = frozenset([7, 8, 9, 11, 12])
    FOLLOW_sig_primary_in_sig_bool_and248 = frozenset([1, 6])
    FOLLOW_NOT_in_sig_primary310 = frozenset([7, 8, 9, 11, 12])
    FOLLOW_sig_primary_in_sig_primary314 = frozenset([1])
    FOLLOW_sig_constant_in_sig_primary356 = frozenset([1])
    FOLLOW_IDENT_in_sig_primary388 = frozenset([1])
    FOLLOW_PG_in_sig_primary429 = frozenset([7, 8, 9, 11, 12])
    FOLLOW_sig_bool1_in_sig_primary434 = frozenset([10])
    FOLLOW_PD_in_sig_primary436 = frozenset([1])
    FOLLOW_T_in_sig_constant491 = frozenset([1])
    FOLLOW_F_in_sig_constant506 = frozenset([1])



def main(argv, stdin=sys.stdin, stdout=sys.stdout, stderr=sys.stderr):
    from antlr3.main import ParserMain
    main = ParserMain("condexpLexer", condexpParser)
    main.stdin = stdin
    main.stdout = stdout
    main.stderr = stderr
    main.execute(argv)


if __name__ == '__main__':
    main(sys.argv)



##
## Filename    : extract_visitor.py
## Author(s)   : Michel Le Borgne
## Created     : 5 nov. 2012
## Revision    :
## Source      :
##
## Copyright 2009 - 2012 : IRISA/IRSET
##
## This library is free software; you can redistribute it and/or modify it
## under the terms of the GNU General Public License as published
## by the Free Software Foundation; either version 2.1 of the License, or
## any later version.
##
## This library is distributed in the hope that it will be useful, but
## WITHOUT ANY WARRANTY, WITHOUT EVEN THE IMPLIED WARRANTY OF
## MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  The software and
## documentation provided hereunder is on an "as is" basis, and IRISA has
## no obligations to provide maintenance, support, updates, enhancements
## or modifications.
## In no event shall IRISA be liable to any party for direct, indirect,
## special, incidental or consequential damages, including lost profits,
## arising out of the use of this software and its documentation, even if
## IRISA have been advised of the possibility of such damage.  See
## the GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this library; if not, write to the Free Software Foundation,
## Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
##
## The original code contained here was initially developed by:
##
##     Michel Le Borgne.
##     IRISA
##     Symbiose team
##     IRISA  Campus de Beaulieu
##     35042 RENNES Cedex, FRANCE
##
##
##     http:
##     mailto:
##
## Contributor(s): Geoffroy Andrieux
##
"""
Test Visitors for extracting submodels.
PathExtractor : abstract a trajectory as a set of path in the transition graph
"""

import unittest
import sys
from cadbiom.antlr3 import ANTLRFileStream, CommonTokenStream

from cadbiom.models.guard_transitions.translators.chart_xml_pid import PidXmlParser
from cadbiom.models.guard_transitions.translators.chart_lang import LangVisitor
#from models.guard_transitions.chart_model import ChartModel

#from models.guard_transitions.translators.chart_xml import \
#                                     MakeModelFromXmlFile
from cadbiom.models.guard_transitions.extract_visitor import PathExtractor
from cadbiom.models.guard_transitions.analyser.ana_visitors import EstimExpVisitor, \
                       TableVisitor,  FrontierVisitor, SigExpIdCollectVisitor
from cadbiom.models.guard_transitions.translators.cadlangLexer import cadlangLexer
from cadbiom.models.guard_transitions.translators.cadlangParser import cadlangParser
from cadbiom.models.guard_transitions.chart_model import ChartModel

class Reporter(object):
    """
    Simple reporter for tests
    """
    def __init__(self):
        self.error = False
        self.mess = ""

    def display(self,  mes):
        """
        print message
        """
        self.error = True
        mess = self.mess + "ERROR -> "+mes
        print "\n\n"+ mess

class TestExtractVisitors(unittest.TestCase):
    """
    Test visitor for model extraction
    """
    def test_exec1(self):
        """
        Test on simple exp
        """
        file_name = 'translators/tests/ext1.cal'
        reporter = Reporter()
        readf = ANTLRFileStream(file_name)
        lexer = cadlangLexer(readf)
        lexer.set_error_reporter(reporter)
        parser = cadlangParser(CommonTokenStream(lexer))
        parser.set_error_reporter(reporter)
        model = ChartModel('test1')
        parser.cad_model(model)
        model = parser.model

        target = ['C_D']
        frontier = ['F1', 'F2','F3']
        pex = PathExtractor(model, target, frontier, reporter)
        model.accept(pex)
        # decompiler visitor
        lvi = LangVisitor(sys.stdout)
        pex.extract_model.accept(lvi)

#    def test_exec2(self):
#        """
#        test visitor execution
#        """
#        file_name = 'translators/tests/tgf_cano.xml'
#        reporter = Reporter()
#        parser = PidXmlParser(file_name, reporter)
#        if reporter.error:
#            print 'ERROR IN READING FILE'
#        model = parser.model
#
#        tvi = TableVisitor(reporter)
#        model.accept(tvi)
#
#        target = ['TRAP_1']
#        frontier = ['TGFBR2__dimer___active_intToMb',
#                    'TGFBfamily__dimer___active_exCellRegion', 'CTGF']
#        pex = PathExtractor(model, target, frontier, reporter)
#        model.accept(pex)
#        # decompiler visitor
#        lvi = LangVisitor(sys.stdout)
#        pex.extract_model.accept(lvi)



## Filename    : static_analysis.py
## Author(s)   : Michel Le Borgne
## Created     : 02/2012
## Revision    :
## Source      :
##
## Copyright 2012 : IRISA/IRSET
##
## This library is free software; you can redistribute it and/or modify it
## under the terms of the GNU General Public License as published
## by the Free Software Foundation; either version 2.1 of the License, or
## any later version.
##
## This library is distributed in the hope that it will be useful, but
## WITHOUT ANY WARRANTY, WITHOUT EVEN THE IMPLIED WARRANTY OF
## MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  The software and
## documentation provided here under is on an "as is" basis, and IRISA has
## no obligations to provide maintenance, support, updates, enhancements
## or modifications.
## In no event shall IRISA be liable to any party for direct, indirect,
## special, incidental or consequential damages, including lost profits,
## arising out of the use of this software and its documentation, even if
## IRISA have been advised of the possibility of such damage.  See
## the GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this library; if not, write to the Free Software Foundation,
## Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
##
## The original code contained here was initially developed by:
##
##     Michel Le Borgne.
##     IRISA/IRISA
##     Symbiose team
##     IRISA  Campus de Beaulieu
##     35042 RENNES Cedex, FRANCE
##
##
## Contributor(s): Geoffroy Andrieux
##
"""
Classes for static analysis
"""
import string
import networkx as nx
import matplotlib.pyplot as plt
import copy
from collections import OrderedDict

from cadbiom.antlr3 import ANTLRFileStream, CommonTokenStream


from cadbiom.models.biosignal.translators.gt_visitors import compile_cond, compile_event
from cadbiom.models.guard_transitions.translators.cadlangLexer import cadlangLexer
from cadbiom.models.guard_transitions.translators.cadlangParser import cadlangParser
from cadbiom.models.guard_transitions.translators.chart_xml import MakeModelFromXmlFile
from cadbiom.models.guard_transitions.translators.chart_xml_pid import \
                                          MakeModelFromPidFile
from cadbiom.models.guard_transitions.chart_model import ChartModel
from cadbiom.models.guard_transitions.analyser.ana_visitors import EstimExpVisitor, \
                       TableVisitor,  FrontierVisitor, SigExpIdCollectVisitor
import cadbiom.commons as cm

LOGGER = cm.logger()

class StaticAnalyzer(object):
    """
    Main class for static analysis
    """

    def __init__(self, reporter):
        self.reporter = reporter
        self.model = None
        self.influence_dict = dict()

    def build_from_chart_model(self, chart_model):
        """
        For technical purpose
        """
        self.model = chart_model

    def build_from_cadlang(self, file_name):
        """
        Build a StaticAnalyser from a .cal file of PID database
        @param file_name: str - path of .cal file
        """

        crep = self.reporter
        fstream = ANTLRFileStream(file_name)
        lexer = cadlangLexer(fstream)
        lexer.set_error_reporter(crep)
        parser = cadlangParser(CommonTokenStream(lexer))
        parser.set_error_reporter(crep)
        model = ChartModel(file_name)
        parser.cad_model(model)
        # chart model
        self.model = parser.model

    def build_from_chart_file(self, file_name):
        """
        Build a StaticAnalyser from a .bcx file
        @param file_name: str - path of the .bcx file
        """
        parsing = MakeModelFromXmlFile(file_name)
        # chart model
        self.model = parsing.get_model()

    def build_from_pid_file(self, file_name, has_clock=True,
                            ai_interpretation=0):
        """
        Build an StaticAnalyser from a .xml file of PID database
        @param file_name: str - path of .xml file
        """
        parser = MakeModelFromPidFile(file_name, self.reporter,
                                       has_clock, ai_interpretation)
        # chart model
        self.model = parser.get_model()

    def get_frontier_node_names(self):
        """
        returns a (list<string>) list of frontier place names
        """
        fvi = FrontierVisitor()
        self.model.accept(fvi)
        return fvi.frontier

    def get_gene_places(self):
        """
        We assume that all places representing a gene contain the 'gene' string in the name
        """
        lpname = []
        node_dict = self.model.node_dict
        for pname in node_dict.keys():
            if string.find(pname, 'gene') > 0:
                lpname.append(pname)
        return lpname


    def get_statistics(self, sfile=False):
        """
        Compute various informations on the model (assume PID encoding)

        .. warning:: Start nodes (__start__0) are counted here in 'other'
            locations.

        :param arg1: Opened file (optional).
        :type arg1: <open file>
        :return: Informations (status of nodes and their cellular locations).
        :rtype: <str>
        """

        input_cpt = 0
        front_cpt = 0
        final_cpt = 0
        isolated_cpt = 0
        locations = (
            '_gene', '_transMb', '_nucl', '_intToMb', '_cy', '_plasmaMb',
            '_exCellRegion', '_ccJct', '_en', '_siteofdouble_strandbreak',
            '_membraneraft', 'other',
        )
        # Init counter of locations
        location_dict = OrderedDict({location: 0 for location in locations})

        # Iterate on model nodes
        for node_name, node in self.model.node_dict.iteritems():
            loc_found = False
            for location in locations:
                if location in node_name:
                    if loc_found:
                        LOGGER.warning(
                            "get_statistics::Multiple locations: {}, {}".format(
                                node_name, location))
                    location_dict[location] += 1
                    loc_found = True


            # Warning: this line also counts "virtual" "start nodes"
            # like __start__0
            if not loc_found:
                location_dict['other'] += 1

            # Check node status
            nb_outgoing_trans = len(node.outgoing_trans)
            nb_incoming_trans = len(node.incoming_trans)

            if node.is_input():
                input_cpt += 1
            if nb_outgoing_trans == 0:
                final_cpt += 1
            if nb_incoming_trans == 0:
                front_cpt += 1
            if nb_outgoing_trans == 0 and nb_incoming_trans == 0:
                isolated_cpt += 1

        # Forge data header
        data = OrderedDict({
            "Model name": self.model.name,
            "Nb places": len(self.model.node_dict),
            "Nb transitions": len(self.model.transition_list),
            "Nb inputs": input_cpt,
            "Nb frontier (no incoming transitions)": front_cpt,
            "Nb terminal (no outgoing transitions)": final_cpt,
            "Nb isolated (frontier places in conditions/not used)":isolated_cpt,
        })
        # Forge data locations
        for k, v in location_dict.iteritems():
            data['Nb ' + k] = v

        # Forge text
        text = "\n".join([k + ': ' + str(v) for k, v in data.iteritems()])

        # Dump text if an opened file is given
        if sfile != False:
            sfile.write(text)

        return text


    def get_frontier_scc(self):
        """
        Find initial strongly connected components in the transition graph
        @return:   list<list<string>> - list of list of node names in frontier
        strongly connected components

        .. warning:: If the model is modified but not reloaded, KeyError is
            raised (see below); we inform the user of that...
        """
        # Get directed graph
        directed_graph = self.make_transition_dg()
        # generator on the strongly connectec components
        scc_list = nx.strongly_connected_components(directed_graph)

        frontier_list = []
        tvi = TableVisitor(self.reporter)
        self.model.accept(tvi)

        for scc in scc_list:
            # eliminate isolated frontier nodes
            if len(scc) > 1:
                frontier = True
                # generate symbol table for partial evaluation
                # all places in scc are inactivated (-1)
                symbol_table = dict.fromkeys(scc, -1)
                pe_visitor = EstimExpVisitor(symbol_table)

                # find all incoming transitions for each node in scc
                LOGGER.debug("scc: " + str(scc))
                for node_name in scc:

                    try:
                        # raises keyerror if the model isn't reloaded manually..
                        chart_node = self.model.node_dict[node_name]
                    except KeyError:
                        return list([[
                            "Error: Please try by \
                            saving & reloading the model manually"]])

                    for in_trans in chart_node.incoming_trans:
                        LOGGER.debug(
                            "Node {}, trans: {}".format(
                                node_name,
                                (in_trans.ori.name, in_trans.ext.name)
                            )
                        )

                        # if the transition comes from outside scc,
                        # we try to detect if scc may be reached with this trans
                        if in_trans.ori.name not in scc :
                            cond = self.__tr_estim(in_trans, pe_visitor,
                                                   tvi.tab_symb)

                            LOGGER.debug(
                                "scc validated as a cyle ?: " + str(cond)
                            )
                            # Guard of the transition is True
                            # (not False or indeterminate),
                            # so scc may be reached from other part of the model
                            # => switch to next scc
                            if cond >= 0:
                                frontier = False
                                break
                if frontier:
                    LOGGER.debug("frontier found")
                    frontier_list.append(scc)

        LOGGER.debug("SCC found: " + str(frontier_list))

        # Nothing found => return empty result
        if len(frontier_list) == 0:
            return [[]]
        return frontier_list

    def get_basal_activated_genes(self):
        """
        Basal activated genes are genes which are activated even if all the places present
        in activation condition, are inactivate
        """
        node_dict = self.model.node_dict
        tvi = TableVisitor(self.reporter)
        self.model.accept(tvi)
        lg_n = self.get_gene_places()
        lgene = []
        for gen in lg_n:
            lgene.append(node_dict[gen])
        lbag = []

        for gep in lgene:
            pe_visitor = EstimExpVisitor(None) # all variables set to False
            assert(len(gep.outgoing_trans)==1)
            trans = gep.outgoing_trans[0] # assume only one transition
            trans_val = self.__tr_estim(trans, pe_visitor, tvi.tab_symb)
            if trans_val == 1:
                lbag.append(gep.name)
        lbag.sort()
        return lbag

    def get_why_basal_genes(self):
        """
        More complete than get_basal_activated_genes. Returns a list of pairs (g,lpv) where
        g is a basal activated gene and lpv are the principal variables (isolated_inhib., dominant_inhib,
        essential_act, dominant activator)
        """
        res = []
        node_dict = self.model.node_dict
        lbag = self.get_basal_activated_genes()
        for gep in lbag:
            gnode = node_dict[gep]
            trans = gnode.outgoing_trans[0]
            pva = self.get_tr_principal_variables(trans)
            res.append((gep, pva))
        return res


    def __tr_estim(self, trans, e_visitor, tab_symb):
        """
        Given a partial assignation of variables implemented in tab_symb, estimate if
        the guard of a transition is True, False or indeterminate. The variables not present
        in the symbol table are considered as indeterminates

        If tab_symb is None, all variables are assigned the value False and the condition of the
        guard is estimated alone.
        """
        ev_estim = 1
        cond_estim = 1

        if trans.condition:
            cond_sexpr = compile_cond(trans.condition, tab_symb, self.reporter)
            cond_estim = cond_sexpr.accept(e_visitor)

        if not e_visitor.symb_table:
            return cond_estim

        if trans.event:
            ev_sexpr = compile_event(trans.event, tab_symb,
                                        True, self.reporter)[0]
            ev_estim = ev_sexpr.accept(e_visitor)

        return min(ev_estim, cond_estim)

    def __get_tr_node_variables(self, cond_sexpr, ev_sexpr):
        """
        Get the node idents in the guard expression of
        the transition condition and event
        """
        lst1 = []
        lst2 = []
        lcl = []
        icv = SigExpIdCollectVisitor()
        if cond_sexpr:
            # condition expressions contains only node ident
            lst1 = cond_sexpr.accept(icv)
        if ev_sexpr:
            ll2 = ev_sexpr.accept(icv)
            for ident in ll2:
                if self.model.node_dict.has_key(ident): # state
                    lst2.append(ident)
                else: # clock
                    lcl.append(ident)
        lst = lst1 + lst2
        lst.sort()
        # eliminate doubles
        if len(lst)>0:
            nres = [lst[0]]
            for i in range(len(lst)-1):
                if lst[i+1] != lst[i]:
                    nres.append(lst[i+1])
        else:
            nres = []

        if len(lcl)>0:
            cres = [lcl[0]]
            for i in range(len(lcl)-1):
                if lcl[i+1] != lcl[i]:
                    cres.append(lcl[i+1])
        else:
            cres = []

        return nres , cres

    # for test
    def test_get_tr_node_variables(self, trans):
        """
        externalize a private method for test only
        """
        tvi = TableVisitor(self.reporter)
        self.model.accept(tvi)

        if trans.condition:
            cond_sexpr = compile_cond(trans.condition, tvi.tab_symb,
                                      self.reporter)
        else:
            cond_sexpr = None
        if trans.event:
            ev_sexpr, see, free_clocks = compile_event(trans.event,
                                                       tvi.tab_symb,
                                                       True, self.reporter)
        else:
            ev_sexpr = None

        lvar = self.__get_tr_node_variables(cond_sexpr, ev_sexpr)
        return lvar


    def get_tr_principal_variables(self, trans):
        """
        Compute the essentiel, inhibitors, dominant inhibitors,
        essential activators and dominant activators of a transition tr.
        """
        ein = [] # essential inhibitors
        din = [] # dominant inhibitors
        eac = [] # essentiel activators
        dac = [] # dominant activators
        tvi = TableVisitor(self.reporter)
        self.model.accept(tvi)
        # compile guard expression
        if trans.condition:
            cond_sexpr = compile_cond(trans.condition, tvi.tab_symb,
                                      self.reporter)
        else:
            cond_sexpr = None
        if trans.event:
            ev_sexpr, see, free_clocks = compile_event(trans.event,
                                                      tvi.tab_symb,
                                                      True, self.reporter)
        else:
            ev_sexpr = None

        lvar , lcl = self.__get_tr_node_variables(cond_sexpr, ev_sexpr)

        for ident in lvar:
            # essential inhibitor (F => (guard=T))
            assign_table = dict()
            assign_table[ident] = -1
            for clo in lcl:
                assign_table[clo] = 1 # clocks are assigned to Present
            pe_visitor = EstimExpVisitor(assign_table)

            if cond_sexpr:
                cpe = cond_sexpr.accept(pe_visitor)
            else:
                cpe = 1
            if ev_sexpr:
                epe = ev_sexpr.accept(pe_visitor)
            else:
                epe = 1
            if min(cpe, epe) == 1:
                ein.append(ident)
            # dominant inhibitor (T => (guard=F))
            assign_table = dict()
            assign_table[ident] = 1
            for clo in lcl:
                assign_table[clo] = 1 # clocks are assigned to Present
            pe_visitor = EstimExpVisitor(assign_table)

            if cond_sexpr:
                cpe = cond_sexpr.accept(pe_visitor)
            else:
                cpe = 1
            if ev_sexpr:
                epe = ev_sexpr.accept(pe_visitor)
            else:
                epe = 1
            if min(cpe, epe) == -1:
                din.append(ident)
            # essential activator (F => (guard=F))
            assign_table = dict()
            assign_table[ident] = -1
            for clo in lcl:
                assign_table[clo] = 1 # clocks are assigned to Present
            pe_visitor = EstimExpVisitor(assign_table)

            if cond_sexpr:
                cpe = cond_sexpr.accept(pe_visitor)
            else:
                cpe = 1
            if ev_sexpr:
                epe = ev_sexpr.accept(pe_visitor)
            else:
                epe = 1
            if min(cpe, epe) == -1:
                eac.append(ident)
            # dominant activator (T => (guard=T))
            assign_table = dict()
            assign_table[ident] = 1
            for clo in lcl:
                assign_table[clo] = 1 # clocks are assigned to Present
            pe_visitor = EstimExpVisitor(assign_table)

            if cond_sexpr:
                cpe = cond_sexpr.accept(pe_visitor)
            else:
                cpe = 1
            if ev_sexpr:
                epe = ev_sexpr.accept(pe_visitor)
            else:
                epe = 1
            if min(cpe, epe) == 1:
                dac.append(ident)
        return [ein, din, eac, dac]


    def make_transition_dg(self):
        """
        Build the directed graph of transitions (flow graph)
        @return: a Networkx digraph
        """
        graph = nx.DiGraph()

        edges = [(trans.ori.name, trans.ext.name)
            for trans in self.model.transition_list]

        graph.add_edges_from(edges)
        graph.add_nodes_from(self.model.node_dict.keys())

        return graph

    def make_dependence_dg(self, weight = False):
        """
        Build the directed graph of transitions and conditions (dependence graph)
        if weight : dependencies from transitions have a weight of 2, dependencies from conditions have a weight of 1
        else : all dependencies have a weight of 1
        @return: a Networkx digraph
        """
        graph = nx.DiGraph()
        for trans in self.model.transition_list :
            ori_name = trans.ori.name
            ext_name = trans.ext.name

            lno , lcl = self.test_get_tr_node_variables(trans)
            for cond in lno :
                graph.add_edges_from([(cond, ext_name)], weight=1)
            if weight :
                graph.add_edges_from([(ori_name, ext_name)], weight=2)
            else :
                graph.add_edges_from([(ori_name, ext_name)], weight=1)
        return graph


    def make_full_dependence_dg(self, weight = False):
        """
        Build the directed graph of transitions and conditions (dependence graph)
        Conditions also influence input
        @return: a Networkx digraph
        """
        graph = nx.DiGraph()
        for trans in self.model.transition_list :
            ori_name = trans.ori.name
            ext_name = trans.ext.name

            lno , lcl = self.test_get_tr_node_variables(trans)
            for cond in lno :
                graph.add_edges_from([(cond, ext_name)], weight=1)
                # conditions influence input
                graph.add_edges_from([(cond, ori_name)], weight=1)

            if weight :
                graph.add_edges_from([(ori_name, ext_name)], weight=2)
            else :
                graph.add_edges_from([(ori_name, ext_name)], weight=1)
        return graph

#    def make_full_dependence_weighted_dg(self):
#        """
#        Build the directed graph of transitions and conditions
#        (dependence graph)
#        Dependences from transitions have a weight of 2
#        Dependences from conditions have a weight of 1
#        Conditions also influence input
#        @return: a Networkx digraph
#        """
#        G=nx.DiGraph()
#        for trans in self.model.transition_list :
#            ori_name = trans.ori.name
#            ext_name = trans.ext.name
#
#            ln , lh = self.test_get_tr_node_variables(trans)
#            for cond in ln :
#                G.add_edges_from([(cond,ext_name)], weight=1)
#                # conditions influence input
#                G.add_edges_from([(cond,ori_name)], weight=1)
#
#            G.add_edges_from([(ori_name,ext_name)], weight=2)
#        return G

#    def condition_count(self):
#        cond_dict=dict()
#        for trans in self.model.transition_list :
#            if trans.condition != '':
#                cond_name_list = get_property_places(trans.condition)
#                for cond in cond_name_list :
#                    if cond_dict.has_key(cond) :
#                        cond_dict[cond]+=1
#                    else :
#                        cond_dict[cond]=1
#        cond_items = cond_dict.items()
#        cond_items.sort(cmpval)
#
#        cond_moy=0.0
#        for cond_couple in cond_items :
#            cond_moy+=cond_couple[1]
#        print 'cond moy : ',cond_moy/len(cond_dict.keys())
#        print 'cond median : ',cond_items[len(cond_dict.keys())/2][1]
#        print 'cond count : ',cond_items[:10]
#        return cond_items

    def get_predecessors(self, dgraph, node, rank=None):
        """
        return all node predecessors if rank=None, predecessors under the rank else
        @return: a list of str predecessor name
        """
        if not rank :
            pred_list = dgraph.predecessors(node)
            pred_list.append(node)
            for pred in pred_list:
                sec_pred_list = dgraph.predecessors(pred)
                for sec_p in sec_pred_list:
                    if sec_p not in pred_list:
                        pred_list.append(sec_p)
#            print len(pred_list), pred_list
        else :
            self.ranked_list = []
            pred_list = [node]
            already_visited = []
            for i in range(rank):
                new_pred_list = []
                for node in pred_list:
                    if node not in already_visited:
                        new_pred_list.extend(dgraph.predecessors(node))
                        already_visited.append(node)
                self.ranked_list.append(new_pred_list)
                pred_list.extend(new_pred_list)
#                print i, len(list(set(new_pred_list)))
#                print i,len(new_pred_list),new_pred_list
        return list(set(pred_list))

    def get_successors(self, dgraph, node, rank=None):
        """
        return all node successors if rank=None, successors under the rank else
        @return: a list of str successor name
        """
        if not rank :
            suc_list = dgraph.successors(node)
            suc_list.append(node)
            for suc in suc_list:
                sec_suc_list = dgraph.successors(suc)
                for sec_s in sec_suc_list:
                    if sec_s not in suc_list:
                        suc_list.append(sec_s)
    #        print len(pred_list), pred_list
        else :
            self.ranked_list = []
            suc_list = [node]
            already_visited = []
            for i in range(rank):
                new_suc_list = []
                for node in suc_list:
                    if node not in already_visited:
                        new_suc_list.extend(dgraph.successors(node))
                        already_visited.append(node)
                self.ranked_list.append(new_suc_list)
                suc_list.extend(new_suc_list)
#                print i,len(new_pred_list),new_pred_list
        return list(set(suc_list))


    def get_local_neighbors(self, dgraph, node):
        """
        return node direct neighbors (successors and predecessors) in a same list
        @return: a list of str neighbor name
        """
        neighbor_list = []
        neighbor_list.extend(dgraph.predecessors(node))
        neighbor_list.extend(dgraph.successors(node))
        return list(set(neighbor_list))

#    def get_list_neighbors(self, dgraph, node_list):
#        neighbor_list = []
#        for node in node_list :
#            neighbor_list.extend(self.get_node_neighbors(dgraph, node))
#        final_list = []
#        final_list.extend(neighbor_list)
#        final_list.extend(node_list)
#        return list(set(final_list))

#    def change_node_att(self, g, node_list, att, value):
#        for node in node_list:
#            g.node[node].size = value
#        return g

    def get_propagated_places(self, frontier_places, property_places):

        final_list = []
        already_visited = []
        condition = []
        new_frontier = frontier_places

        while len(new_frontier) != 0 :
            save_frontier = new_frontier
            new_frontier = []

            for place in save_frontier :
                node = self.model.node_dict[place]
                for trans in node.outgoing_trans :
                    ext_place = trans.ext.name
                    if  ext_place not in already_visited :
                        already_visited.append(ext_place)
                        new_frontier.append(ext_place)

                    lno , lcl = self.test_get_tr_node_variables(trans)
                    condition.extend(lno)
            final_list.extend(new_frontier)

        new_frontier = property_places

        while len(new_frontier) != 0 :
            save_frontier = new_frontier
            new_frontier = []

            for place in save_frontier :
                node = self.model.node_dict[place]
                for trans in node.incoming_trans :
                    ori_place = trans.ori.name
                    if  ori_place not in already_visited :
                        already_visited.append(ori_place)
                        new_frontier.append(ori_place)

                    lno , lcl = self.test_get_tr_node_variables(trans)
                    condition.extend(lno)
            final_list.extend(new_frontier)

        new_frontier = property_places

        while len(new_frontier) != 0 :
            save_frontier = new_frontier
            new_frontier = []

            for place in save_frontier :
                node = self.model.node_dict[place]
                for trans in node.outgoing_trans :
                    ext_place = trans.ext.name
                    if  ext_place not in already_visited :
                        already_visited.append(ext_place)
                        new_frontier.append(ext_place)

                    lno , lcl = self.test_get_tr_node_variables(trans)
                    condition.extend(lno)
            final_list.extend(new_frontier)


        tr_graph = self.make_transition_dg()
        to_check_for_perm = []
        to_check_for_perm.extend(final_list)
        to_check_for_perm.extend(condition)
        to_check_for_perm.extend(property_places)
        to_check_for_perm = list(set(to_check_for_perm))

        perm_list = []
        for place in to_check_for_perm :
            predecessors = self.get_predecessors(tr_graph, place, None)
            for pred in predecessors :
                if self.model.node_dict[pred].is_perm():
                    if pred not in perm_list :
                        perm_list.append(pred)

        for place in perm_list :
            successors = self.get_successors(tr_graph, place, None)
            final_list.extend(successors)

        final_list = list(set(final_list))
        print len(final_list)
        return final_list


    def get_frontier_successors(self, tr_graph, frontier_places, property_places, perm = True):
        final_list = []
        to_check = frontier_places

        #Add permanent nodes
#        if perm :
#            for node in self.model.node_dict.keys():
#                if self.model.node_dict[node].is_perm():
#                    to_check.append(node)

        final_list.extend(frontier_places)
        final_list.extend(to_check)

        for place in to_check :
            successors = self.get_successors(tr_graph, place, None)
            final_list.extend(successors)

        final_list = list(set(final_list))


        perm_list = []
        important_places = []
        act_graph = self.make_dependence_dg()
        for place in to_check :
            successors = self.get_successors(act_graph, place, None)
            important_places.extend(successors)

        important_places.extend(property_places)
        important_places = list(set(important_places))

        for place in important_places :
            predecessors = self.get_predecessors(tr_graph, place, None)
            for pred in predecessors :
                if self.model.node_dict[pred].is_perm():
                    perm_list.append(pred)
        perm_list = list(set(perm_list))

        for place in perm_list :
            successors = self.get_successors(tr_graph, place, None)
            final_list.extend(successors)

        print len(list(set(final_list)))
        print list(set(final_list))
        return list(set(final_list))

    def get_submodel(self, node_list):
        new_model = copy.deepcopy(self.model)
        new_model_nodes = new_model.node_dict.keys()


    def remove_nodes_not_in_list(self, graph, node_list):
        """
        @return: a Networkx graph (or digraph) without nodes NOT in the input list
        """
        to_remove = graph.nodes()
        for node in to_remove :
            if node not in node_list:
                graph.remove_node(node)
        return graph

    def remove_nodes_in_list(self, graph, node_list):
        """
        @return: a Networkx graph (or digraph) without nodes in the input list
        """
        to_remove = graph.nodes()
        for node in to_remove :
            if node in node_list:
                graph.remove_node(node)
        return graph

    def in_degree_cut_off(self, dgraph, degree):
        """
        @return: a Networkx digraph without nodes with input degree upper than degree parameter
        """
        in_degree = dgraph.in_degree()
        node_list = dgraph.nodes()
        for node in node_list :
            if in_degree[node] > degree :
                dgraph.remove_node(node)
        return dgraph

    def out_degree_cut_off(self, dgraph, degree):
        """
        @return: a Networkx digraph without nodes with output degree upper than degree parameter
        """
        out_degree = dgraph.out_degree()
        node_list = dgraph.nodes()
        for node in node_list :
            if out_degree[node] > degree :
                dgraph.remove_node(node)
        return dgraph


    def export_2_graphml(self, graph, graphml_file):
        """
        As it says
        @param graph: the graph to export
        @param graphml_file: filename
        """
        nx.write_graphml(graph, graphml_file)

    def export_2_dot(self, graph, dot_file):
        """
        As it says
        @param graph: the graph to export
        @param dot_file: filename
        """
        nx.write_dot(graph, dot_file)



def stat_on_graph(graph):
    """
    ???
    """
    print 'number of nodes : ', graph.order()
    print 'number of edges : ', graph.number_of_edges()

    print 'number of connected components : ', \
           nx.number_connected_components(graph)

    if nx.is_connected(graph):
        print 'diameter : ', nx.diameter(graph) # max of shortest length
        print 'average path length : ', nx.average_shortest_path_length(graph)
        print 'density : ', nx.density(graph)
        print 'clustering : ', nx.clustering(graph)

    degree = graph.degree().items()
    degree.sort(cmpval)
    degree_moy = 0.0
    for degree_couple in degree :
        degree_moy += degree_couple[1]
    print 'degree moy : ', degree_moy / graph.order()

    print 'degree median : ', degree[len(degree) / 2][1]
    print 'degree : ', degree[:10]


def stat_on_dgraph(dgraph):
    """
    ???
    """
    print 'number of nodes : ', dgraph.order()
    print 'number of edges : ', dgraph.number_of_edges()
    nscc = nx.number_strongly_connected_components(dgraph)
    print 'number of strongly connected components : ', nscc

    in_degree = dgraph.in_degree().items()
    in_degree.sort(cmpval)
    print 'in degree : ', in_degree[:10]

    out_degree = dgraph.out_degree().items()
    out_degree.sort(cmpval)
    print 'out degree : ', out_degree[:10]


def view_graph(graph):
    """
    draw interaction graph with matplotlib
    """
    pos = nx.spring_layout(graph, iterations=5)
    low = []
    medium = []
    high = []
    hyper_high = []
    for node in graph:
        neighbors = graph.neighbors(node)
        if len(neighbors) <= 2:
            low.append(node)
        elif len(neighbors) > 2 and len(neighbors) <= 5:
            medium.append(node)
        elif len(neighbors) > 5 and len(neighbors) <= 10:
            high.append(node)
        else:
            hyper_high.append(node)

    nx.draw_networkx_nodes(graph, pos, nodelist=low, node_color='blue')
    nx.draw_networkx_nodes(graph, pos, nodelist=medium,
                           node_color='deepskyblue')
    nx.draw_networkx_nodes(graph, pos, nodelist=high, node_color='hotpink')
    nx.draw_networkx_nodes(graph, pos, nodelist=hyper_high,
                           node_color='crimson')
    nx.draw_networkx_edges(graph, pos)
#        nx.draw_networkx_labels(graph,pos)
    plt.show()


def view_weighted_graph(graph, rank):
    """
    ???
    """
    pos = nx.spring_layout(graph, iterations=10)
    edges = graph.edges(data=True)
    elarge = [(u, v) for (u, v, d) in edges if d['weight'] > 1]
    esmall = [(u, v) for (u, v, d) in edges if d['weight'] <= 1]

    # nodes

    nx.draw_networkx_nodes(graph, pos, node_size=100, nodelist=rank[0],
                           node_color='crimson')
    nx.draw_networkx_nodes(graph, pos, node_size=75, nodelist=rank[1],
                           node_color='hotpink')
    nx.draw_networkx_nodes(graph, pos, node_size=50, nodelist=rank[2],
                           node_color='deepskyblue')
    nx.draw_networkx_nodes(graph, pos, node_size=25, nodelist=rank[3],
                           node_color='blue')
    # edges
    nx.draw_networkx_edges(graph, pos, edgelist=elarge,
                        width=0.5, alpha=0.5)
    nx.draw_networkx_edges(graph, pos, edgelist=esmall,
                           width=0.5, alpha=0.5,
                           edge_color='dodgerblue', style='dashed')

    # labels
    nx.draw_networkx_labels(graph, pos, font_size=10, font_family='sans-serif')

    plt.axis('off')
#    plt.savefig("weighted_graph.png") # save as png

    plt.show() # display

def cmpval(xxx, yyy):
    """
    utility
    """
    if xxx[1] > yyy[1]:
        return -1
    elif xxx[1] == yyy[1]:
        return 0
    else:
        return 1



##
## Filename    : extract_visitor.py
## Author(s)   : Michel Le Borgne
## Created     : 5 nov. 2012
## Revision    :
## Source      :
##
## Copyright 2009 - 2012 : IRISA/IRSET
##
## This library is free software; you can redistribute it and/or modify it
## under the terms of the GNU General Public License as published
## by the Free Software Foundation; either version 2.1 of the License, or
## any later version.
##
## This library is distributed in the hope that it will be useful, but
## WITHOUT ANY WARRANTY, WITHOUT EVEN THE IMPLIED WARRANTY OF
## MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  The software and
## documentation provided hereunder is on an "as is" basis, and IRISA has
## no obligations to provide maintenance, support, updates, enhancements
## or modifications.
## In no event shall IRISA be liable to any party for direct, indirect,
## special, incidental or consequential damages, including lost profits,
## arising out of the use of this software and its documentation, even if
## IRISA have been advised of the possibility of such damage.  See
## the GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this library; if not, write to the Free Software Foundation,
## Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
##
## The original code contained here was initially developed by:
##
##     Michel Le Borgne.
##     IRISA
##     Symbiose team
##     IRISA  Campus de Beaulieu
##     35042 RENNES Cedex, FRANCE
##
##
##     http:
##     mailto:
##
## Contributor(s):  Geoffroy Andrieux
##
"""
Visitors for extracting submodels.
PathExtractor : abstract a trajectory as a set of path in the transition graph
"""

from cadbiom.models.guard_transitions.chart_model import ChartModel, CTransition
from cadbiom.models.guard_transitions.analyser.ana_visitors import \
        SigExpIdCollectVisitor
from cadbiom.models.biosignal.translators.gt_visitors import compile_cond, compile_event
from cadbiom.models.guard_transitions.analyser.ana_visitors import TableVisitor

class PathExtractor(object):
    """
    Extract a submodel consisting in all the path from a frontier subset to
    a target represented as a list of places
    The visitor clean and affect simple node attributes:
    activated: treated - incoming paths explored
    was_activated: reach frontier or not - has meaning if activated = True
    """
    def __init__(self, model, target, frontier, reporter):
        """
        Constructor
        @param target: a list of places
        @param frontier: a list of frontier places
        """
        self.model = model
        tvi = TableVisitor(reporter)
        model.accept(tvi)
        self.symb_tab = tvi.tab_symb
        self.found = []
        self.frontier = [] # OK frontier
        self.reporter = reporter
        try:
            for pla in target:
                self.found.append(model.simple_node_dict[pla])
            for pla in frontier:
                self.frontier.append(model.simple_node_dict[pla])
        except KeyError:
            pass

        self.extract_model = ChartModel(model.name + "_extract")

    def visit_chart_model(self, model):
        """
        Do the job
        """
        # set activated and was_activated simple node attributes to False
        model.clean()
        # explore all paths and build the sub model
        while len(self.found) > 0:
            xpl = self.found[0]
            self.found.remove(xpl)
            xpl.accept(self)
        # set to false non reachable places in events and conditions

        # clean
        model.clean()

    def visit_cstart_node(self, node):
        """
        do nothing: see transitions
        """
        return False

    def visit_csimple_node(self, node):
        """
        explore all incoming transitions
        @return: True if frontier or perm,input node reached from node
        """
        # node already visited (use activated attribute)
        if node.activated:
            return node.was_activated  # path to frontier or not
        node.activated = True
        # node without incoming transition (frontier but may not be OK)
        if len(node.incoming_trans) == 0:
            node.was_activated = node in self.frontier
            if node.was_activated and len(node.outgoing_trans) == 0:
                # isolated frontier OK node
                self.extract_model.get_root().add_copy(node)
            return node.was_activated
        # node with incoming transition
        natt = False
        for intr in node.incoming_trans:
            # deep first visit of predecessors
            fatt = intr.ori.accept(self)
            if fatt: # reach OK frontier this way
                # find new nodes in event and cond
                tr_var = extract_state_var(self.model, intr,
                                           self.symb_tab, self.reporter)
                # add nodes in found
                for nvar in tr_var:
                    node = self.model.node_dict[nvar]
                    if node.is_simple():
                        if not node.activated: # not treated
                            if not node in self.found: # not waiting
                                self.found.append(node)
                # add the transition to extracted model
                add_ext_transition(intr, self.extract_model)
                natt = True

        node.was_activated = natt
        return natt

    def visit_cinput_node(self, node):
        """
        An input node is a frontier node always available
        """
        return True

    def visit_cperm_node(self, node):
        """
        A permanent node is an always available frontier node
        """
        return True




# helper functions
def extract_state_var(model, intr, symb_tab, reporter):
    """
    Extract state variable (place names) from condition and
    event of a transition.
    """
    lst2 = []
    lst1 = []
    cond_sexpr = None
    ev_sexpr = None
    if intr.condition:
        cond_sexpr = compile_cond(intr.condition, symb_tab, reporter)
    if intr.event:
        ev_sexpr = compile_event(intr.event, symb_tab, True, reporter)[0]

    icv = SigExpIdCollectVisitor()
    if cond_sexpr:
        # condition expressions contains only node ident
        lst1 = cond_sexpr.accept(icv)
    if ev_sexpr:
        ll2 = ev_sexpr.accept(icv)
        for ident in ll2:
            if model.node_dict.has_key(ident): # state
                lst2.append(ident)
    lst = lst1 + lst2
    lst.sort()
    print 'LISTE', lst
    # eliminate doubles
    if len(lst) > 0:
        nres = [lst[0]]
        for i in range(len(lst)-1):
            if lst[i+1] != lst[i]:
                nres.append(lst[i+1])
    else:
        nres = []
    return nres

def add_ext_transition(trans, extract_model):
    """
    Add a copy of trans in extract_model
    """
    onam = trans.ori.name
    tnam = trans.ext.name
    # find origin and extremity
    try:
        new_ori = extract_model.node_dict[onam]
    except KeyError:
        new_ori = extract_model.get_root().add_copy(trans.ori)
    try:
        new_ext = extract_model.node_dict[tnam]
    except KeyError:
        new_ext = extract_model.get_root().add_copy(trans.ext)

    ntr = extract_model.get_root().add_transition(new_ori, new_ext)
    ntr.set_condition(trans.condition)
    ntr.set_event(trans.event)







## Filename    : TestCLUnfolder.py
## Author(s)   : Michel Le Borgne
## Created     : 13 sept. 2012
## Revision    :
## Source      :
##
## Copyright 2012 : IRISA/IRSET
##
## This library is free software; you can redistribute it and/or modify it
## under the terms of the GNU General Public License as published
## by the Free Software Foundation; either version 2.1 of the License, or
## any later version.
##
## This library is distributed in the hope that it will be useful, but
## WITHOUT ANY WARRANTY, WITHOUT EVEN THE IMPLIED WARRANTY OF
## MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  The software and
## documentation provided here under is on an "as is" basis, and IRISA has
## no obligations to provide maintenance, support, updates, enhancements
## or modifications.
## In no event shall IRISA be liable to any party for direct, indirect,
## special, incidental or consequential damages, including lost profits,
## arising out of the use of this software and its documentation, even if
## IRISA have been advised of the possibility of such damage.  See
## the GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this library; if not, write to the Free Software Foundation,
## Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
##
## The original code contained here was initially developed by:
##
##     Michel Le Borgne.
##     IRISA
##     Symbiose team
##     IRISA  Campus de Beaulieu
##     35042 RENNES Cedex, FRANCE
##
##
## Contributor(s):
##
'''
Unitary Test of the unfolder
'''
import unittest


from cadbiom.models.guard_transitions.chart_model import ChartModel
from cadbiom.models.clause_constraints.CLDynSys import Literal, Clause
from cadbiom.models.guard_transitions.analyser.ana_visitors import TableVisitor
from cadbiom.models.clause_constraints.CLDynSys import CLDynSys
from cadbiom.models.clause_constraints.mcl.MCLTranslators import  GT2Clauses
from cadbiom.models.clause_constraints.mcl.CLUnfolder import CLUnfolder

# helper functions

def string_to_clause(text_clause):
    """
    text_clause is a string such that "a,not b,not c"
    it is transformed in the corresponding clause
    WARNING: strict syntax - no check - for tests
    """
    # split into literals
    clause = Clause([])
    llit = text_clause.split(',')
    # translate each literal
    for lit in llit:
        spl = lit.split()
        if len(spl) == 1: # "a"
            clause.add_lit(Literal(spl[0], True))
        else: # "not a"
            clause.add_lit(Literal(spl[1], False))
    return clause

def num_clause_equal(clause_1, clause_2):
    """
    test DIMACS clauses for equality
    """
    if len(clause_1) != len(clause_2):
        return False
    clause_1.sort()
    clause_2.sort()
    for i in range(len(clause_1)):
        if clause_1[i] != clause_2[i]:
            return False
    return True

def model1():
    """
    A simple ChartModel with two nodes and a transition
    """
    # build dynamical system
    model = ChartModel("Test")
    root = model.get_root()
    node_1 = root.add_simple_node('n1', 0, 0)
    node_2 = root.add_simple_node('n2', 0, 0)
    root.add_transition(node_1, node_2)
    return model

def model2():
    """
    A simple ChartModel with two nodes and a transition
    and a cycle without start
    """
    # build dynamical system
    model = model1()
    root = model.get_root()
    node_3 = root.add_simple_node('n3', 0, 0)
    node_4 = root.add_simple_node('n4', 0, 0)
    root.add_transition(node_3, node_4)
    node_5 = root.add_simple_node('n5', 0, 0)
    root.add_transition(node_4, node_5)
    root.add_transition(node_5, node_3)
    return model

def model3():
    """
    A simple ChartModel with two nodes and a transition
    and a cycle without start
    """
    # build dynamical system
    model = model2()
    root = model.get_root()
    node_s = root.add_start_node('n3', 0, 0)
    node_3 = model.get_simple_node('n3')
    root.add_transition(node_s, node_3)
    return model

def model4():
    """
    Simple chart model with one input and two free clocks
    """
    model = ChartModel("Test")
    root = model.get_root()
    node_1 = root.add_simple_node('n1', 0, 0)
    node_2 = root.add_simple_node('n2', 0, 0)
    tr0 = root.add_transition(node_1, node_2)
    tr0.set_event('hh1')
    tr0.set_condition("n3 or n4")
    node_3 = root.add_simple_node('n3', 0, 0)
    node_4 = root.add_simple_node('n4', 0, 0)
    root.add_transition(node_3, node_4)
    node_5 = root.add_simple_node('n5', 0, 0)
    tr1 = root.add_transition(node_4, node_5)
    tr2 = root.add_transition(node_5, node_3)
    tr1.set_event("hh2")
    tr1.set_condition("n1 and n3")
    node_i =  root.add_input_node('in1', 0, 0)
    tri = root.add_transition(node_i, node_1)
    tri.set_condition("n5")
    return model

def create_unfolder(model):
    """
    returns an unfolder for model
    """
    tvisit = TableVisitor(None) # no error display
    model.accept(tvisit)
    cld = CLDynSys( tvisit.tab_symb, None)
    reporter = ErrorReporter()
    cvisit = GT2Clauses(cld, reporter, True)
    model.accept(cvisit)
    # unfolder
    return CLUnfolder(cld)

# simple reporter: register errors
class ErrorReporter(object):
    """
    Simple reporter for tests
    """
    def __init__(self):
        self.error = False


    def display(self, mess):
        """
        as said
        """
        self.error = True
        #print '\n>> '+mess

class TestCLUnfolder(unittest.TestCase):
    """
    Test public and some private methods (test_method form)
    """
    def test_var_name(self):
        """
        Test of variables uncoding and decoding
        """
        model = model1()
        unfolder = create_unfolder(model)
        # naming and coding variables
        cn1 = unfolder.var_dimacs_code('n1')
        cn2 = unfolder.var_dimacs_code('n2')

        res = unfolder.get_var_name(cn1) == 'n1'
        res = res and (unfolder.get_var_name(cn2) == 'n2')
        self.assert_(res,'Error in variable name 1')

        res = unfolder.get_var_name(7) == 'n1'
        self.assert_(res,'Error in variable name 2')
        res = unfolder.get_var_name(8) == 'n2'
        self.assert_(res,'Error in variable name 3')

        res = unfolder.get_var_indexed_name(7) == 'n1_3'
        self.assert_(res,'Error in variable name 4')
        res = unfolder.get_var_indexed_name(8) == 'n2_3'
        self.assert_(res,'Error in variable name 5')



    def test_frontier(self):
        """
        Test frontier computation and encoding
        """
        # model1
        model = model1()
        unfolder = create_unfolder(model)
        # test frontier: should be n1
        cfr =  unfolder.get_frontier()[0]
        res = unfolder.get_var_name(cfr) == 'n1'
        res = res and (len(unfolder.get_frontier()) == 1)
        self.assert_(res,'Error in frontier: model1')

        # model2 (one cycle without start)
        model = model2()
        unfolder = create_unfolder(model)

        # test frontier: should be n1
        cfr =  unfolder.get_frontier()[0]
        res = unfolder.get_var_name(cfr) == 'n1'
        res = res and (len(unfolder.get_frontier()) == 1)
        self.assert_(res,'Error in frontier: model2')

        # model3: same as model2 but with a start on n3
        model = model3()
        unfolder = create_unfolder(model)

        # test frontier - should be {n1, n3}
        cfr =  unfolder.get_frontier()
        res = len(cfr) == 2
        res = res and unfolder.get_var_name(cfr[0]) == 'n1'
        res = res and unfolder.get_var_name(cfr[1]) == 'n3'
        self.assert_(res,'Error in frontier: model3')

        # model4
        model = model4()
        unfolder = create_unfolder(model)
        cfr =  unfolder.get_frontier()
        res = len(cfr) == 1
        res = res and unfolder.get_var_name(cfr[0]) == 'n3'
        self.assert_(res,'Error in frontier: model4')


    def test_free_clocks_inputs(self):
        """
        test on free clocks and input computation
        """
        # model3: no free clock, no input
        model = model3()
        unfolder = create_unfolder(model)
        lfc = unfolder.get_free_clocks()
        res = len(lfc) == 0
        self.assert_(res,'Error in free clocks: model3')
        lin = unfolder.get_inputs()
        res = len(lin) == 0
        self.assert_(res,'Error in inputs 1')

        # model4: two free clocks and one input
        model = model4()
        unfolder = create_unfolder(model)
        lfc = unfolder.get_free_clocks()
        res = len(lfc) == 2
        res = res and (unfolder.get_var_name(lfc[0]) == 'hh2')
        res = res and (unfolder.get_var_name(lfc[1]) == 'hh1')
        self.assert_(res,'Error in free clocks: model4')
        lin = unfolder.get_inputs()
        res = len(lin) == 1
        res = res and (unfolder.get_var_name(lin[0]) == 'in1')
        self.assert_(res,'Error in inputs: model4')

    def test_init_forward(self):
        """
        test forward initialization for various models and property
        """
        pass



if __name__ == "__main__":
    unittest.main()





# -*- coding: utf-8 -*-
# MIT License
#
# Copyright (c) 2017 IRISA, Jean Coquet, Pierre Vignet
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Contributor(s): Jean Coquet, Pierre Vignet

"""
This module is used to bring some corrections to Cadbiom model processed before.

    - Removing of Strongly Connected Components
    (cycles of disabled places that can never be activated by the solver)

"""
from __future__ import unicode_literals
from __future__ import print_function

# Standard imports
import os

# Custom imports
from cadbiom.models.guard_transitions.analyser.static_analysis \
    import StaticAnalyzer
from cadbiom.models.guard_transitions.translators.chart_xml \
    import XmlVisitor
import cadbiom.commons as cm

LOGGER = cm.logger()


class ErrorRep(object):
    # Cf class CompilReporter(object):
    # gt_gui/utils/reporter.py
    def __init__(self):
        self.context = ""
        self.error = False

    def display(self, mess):
        self.error = True
        LOGGER.error(">> Context: {}; {}".format(self.context, mess))
        exit()

    def display_info(self, mess):
        LOGGER.error("-- Context: {}; {}".format(self.context, mess))
        exit()

    def set_context(self, cont):
        self.context = cont


def add_start_nodes(filePath):
    """Handle Strongly Connected Components (SCC) by adding Start Nodes

    .. note:: Only 1 start node in each SCC is sufficient to suppress it
        from the model.
    .. note:: We use cadbiom API to add Start Nodes and write a new model.

    .. note:: Save the model with "_without_scc" suffix in filename

    :param: File path.
    :type: <str>
    :return: The model loaded by StaticAnalyzer
    :rtype: <ChartModel>
    """

    # Build StaticAnalyzer with Error Reporter
    staticanalyser = StaticAnalyzer(ErrorRep())
    staticanalyser.build_from_chart_file(filePath)
    # Get Strongly Connected Components
    sccs = staticanalyser.get_frontier_scc()

    LOGGER.info("{} SCC found: {}".format(len(sccs), sccs))
    LOGGER.info("Before adding start nodes: " + staticanalyser.get_statistics())

    # Lexicographic sort of nodes in each Strongly Connected Components
    g = (scc for scc in sccs if len(scc) != 0)
    for scc in g:
        scc.sort(key=str.lower)
        # Mark the first node as a frontier
        LOGGER.debug("SCC {}; first lexicographic node selected:{}".format(
                scc, scc[0]))
        staticanalyser.model.mark_as_frontier(scc[0])

    # Save the model with "_without_scc" suffix in filename
    xml = XmlVisitor(staticanalyser.model)
    filename, file_extension = os.path.splitext(filePath)
    mfile = open(filename + "_without_scc" + file_extension, 'w')
    mfile.write(xml.return_xml())
    mfile.close()

    return staticanalyser.model



## Filename    : chart_xml.py
## Author(s)   : Geoffroy Andrieux
## Created     : 04/2010
## Revision    :
## Source      :
##
## Copyright 2010 : IRISA
##
## This library is free software; you can redistribute it and/or modify it
## under the terms of the GNU General Public License as published
## by the Free Software Foundation; either version 2.1 of the License, or
## any later version.
##
## This library is distributed in the hope that it will be useful, but
## WITHOUT ANY WARRANTY, WITHOUT EVEN THE IMPLIED WARRANTY OF
## MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  The software and
## documentation provided here under is on an "as is" basis, and IRISA has
## no obligations to provide maintenance, support, updates, enhancements
## or modifications.
## In no event shall IRISA be liable to any party for direct, indirect,
## special, incidental or consequential damages, including lost profits,
## arising out of the use of this software and its documentation, even if
## IRISA have been advised of the possibility of such damage.  See
## the GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this library; if not, write to the Free Software Foundation,
## Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
##
## The original code contained here was initially developed by:
##
##     Geoffroy Andrieux.
##     IRISA/IRSET
##     Symbiose team
##     IRISA  Campus de Beaulieu
##     35042 RENNES Cedex, FRANCE
##
##
## Contributor(s): Michel Le Borgne, Nolwenn Le Meur
##
"""
Load and generate Cadbiom xml files
"""

from cadbiom.models.guard_transitions.chart_model import ChartModel
from xml.sax import make_parser
from xml.sax import parseString as PS
from xml.sax.handler import ContentHandler
from lxml import etree
from lxml import objectify


class XmlVisitor:
    """
    Visitor used to generate xml cadbiom code
    """

    def __init__(self, model):
        self.model_name = model.name
        self.model = model
        self.fact_list = []
        self.xml = ""      # string: xml representation of model
        self.symb = dict() # symbol table to check double naming of nodes
        self.visit_chart_model()

    def visit_chart_model(self):
        """
        Entrance point
        """
        self.visit_ctop_node(self.model.get_root())

#CNode:
#    def write_xml(self):
#        pass

    def check_name(self, name):
        """
        Detect double declarations
        """
        try:
            dec = self.symb[name]
        except:
            self.symb[name] = "ok"
            return
        raise XmlException("Node double declaration")

    def visit_cstart_node(self, snode):
        """
        Generate xml representation of a start node
        """

        tag = "CStartNode"
        attrname = ["name", "xloc", "yloc"]

        attr = [snode.name, snode.xloc, snode.yloc]
        return [tag, attrname, attr]

    def visit_ctrap_node(self, tnode):
        """
        Generate xml representation of a trap node
        """
        tag = "CTrapNode"
        attrname = ["name", "xloc", "yloc"]
        attr = [tnode.name, tnode.xloc, tnode.yloc]
        return [tag, attrname, attr]

    def visit_csimple_node(self, sin):
        """
        Generate xml representation of a simple node
        """
        self.check_name(sin.name)
        tag = "CSimpleNode"
        attrname = ["name", "xloc", "yloc"]
        attr = [sin.name, sin.xloc, sin.yloc]
        return [tag, attrname, attr]

    def visit_cperm_node(self, pnode):
        """
        Generate xml representation of a perm node
        """
        self.check_name(pnode.name)
        tag = "CPermNode"
        attrname = ["name", "xloc", "yloc"]
        attr = [pnode.name, pnode.xloc, pnode.yloc]
        return [tag, attrname, attr]

    def visit_cinput_node(self, inn):
        """
        Generate xml representation of an input node
        """
        # double declaration of input nodes is allowed"
        tag = "CInputNode"
        attrname = ["name", "xloc", "yloc"]
        attr = [inn.name, inn.xloc, inn.yloc]
        return [tag, attrname, attr]


    def visit_cmacro_node(self, mnode):
        """
        Generate xml representation of a macro node
        """
        self.check_name(mnode.name)
        save_macro = self.current_element
        tag = "CMacroNode"
        attrname = ["name", "xloc", "yloc", "wloc", "hloc"]
        attr = [mnode.name, mnode.xloc, mnode.yloc, mnode.wloc, mnode.hloc]
        properties = [tag, attrname, attr]

        macro = etree.SubElement(self.current_element, properties[0])
        self.current_element = macro
        if len(properties) > 1:
            attrname = properties[1]
            attr = properties[2]
            attributes = macro.attrib
            for i in range(0, len(attrname)):
                attributes[attrname[i]] =  str(attr[i])

        # nodes
        for snode in mnode.sub_nodes:
            properties = snode.accept(self)
            if properties[0] == 'CMacroNode':
                self.current_element = macro

            if properties[0] != 'CMacroNode':
                subel = etree.SubElement(self.current_element, properties[0])
                if len(properties) > 1:
                    attrname = properties[1]
                    attr = properties[2]
                    attributes = subel.attrib
                    for i in range(0, len(attrname)):
                        attributes[attrname[i]] =  str(attr[i])

        # transitions
        for gtr in mnode.transitions:
            for trunlist in gtr:
                properties = trunlist.accept(self)
                sub_tr = etree.SubElement(self.current_element, properties[0])
                if len(properties) > 1:
                    attrname = properties[1]
                    attr = properties[2]
                    attributes = sub_tr.attrib
                    for i in range(0, len(attrname)):
                        attributes[attrname[i]] =  str(attr[i])
        self.current_element = save_macro
        return [tag, attrname, attr]

    def visit_ctop_node(self, tnode):
        """
        interative build of xml tree for model saving
        """
        header = objectify.ElementMaker(annotate=False,
                                        namespace="http://cadbiom",
                                        nsmap={None : "http://cadbiom"})
        xmodel = header.model(name=self.model_name)
        # nodes
        self.current_element = xmodel
        for snode in tnode.sub_nodes:
            properties = snode.accept(self)
            if properties[0] != 'CMacroNode':
                subel = etree.SubElement(xmodel, properties[0])
                if len(properties) > 1:
                    attrname = properties[1]
                    attr = properties[2]
                    attributes = subel.attrib
                    for i in range(0, len(attrname)):
                        attributes[attrname[i]] =  str(attr[i])

        # transitions
        for gtr in tnode.transitions:
            for trans in gtr:
                properties = trans.accept(self)
                sub_tr = etree.Element(properties[0])
                if len(properties) > 1:
                    attrname = properties[1]
                    attr = properties[2]
                    attributes = sub_tr.attrib
                    for i in range(0, len(attrname)):
                        attributes[attrname[i]] =  str(attr[i])
                if trans.note:
                    sub_tr.text = trans.note
                xmodel.append(sub_tr)

        # constraints
        if len(tnode.model.constraints) > 0:
            const = etree.Element("constraints")
            const.text = tnode.model.constraints
            xmodel.append(const)

        self.xml = etree.tostring(xmodel, pretty_print=True)
#        print (etree.tostring(xmodel,pretty_print=True))


    def visit_ctransition(self, trans):
        """
        Generate xml representation of a transition
        """
        tag = "transition"
        attrname = ["name", "ori", "ext", "event",
                    "condition", "action", "fact_ids"]
        attr = [trans.name, trans.ori.name, trans.ext.name, trans.event,
                trans.condition, trans.action, trans.fact_ids]

        fact_ids = trans.fact_ids
        for fact in fact_ids:
            self.fact_list.append(fact)

        return [tag, attrname, attr]

    def return_xml(self):
        """
        get xml string
        """
        return self.xml

    def get_fact_ids(self):
        """
        get litterature references
        """
        model_fact = []
        for i in self.fact_list:
            if i in model_fact:
                continue
            else :
                model_fact.append(i)
        return model_fact


class MakeHandler(ContentHandler):
    """
    make a handler for the parser
    """

    def __init__(self, model = None):
        self.pile_node = []
        self.top_pile = None
        self.pile_dict = []
        self.node_dict = dict()
        self.in_constraints = False
        self.in_transition = False
        self.constraints = ""
        self.model = model

    def startElement(self, name, att):
        if name == "model":
            if not self.model:
                self.model = ChartModel(att.get('name',''))
            root = self.model.get_root()
            self.pile_node.append(root)
            self.top_pile = root
            new_dict = dict()
            self.pile_dict.append(new_dict)
            self.node_dict = new_dict

        elif name == "CStartNode":
            name = att.get('name', '').encode('ascii')
            xloc = float(att.get('xloc', ''))
            yloc = float(att.get('yloc', ''))
            node = self.top_pile.add_start_node(xloc, yloc, name)
            self.node_dict[name] = node

        elif name == "CTrapNode":
            name = att.get('name', '').encode('ascii')
            xloc = float(att.get('xloc', '') )
            yloc = float(att.get('yloc', '') )
            node = self.top_pile.add_trap_node(xloc, yloc, name)
            self.node_dict[name] = node

        elif name == "CSimpleNode":
            name = att.get('name', '').encode('ascii')
            xloc = float(att.get('xloc', '') )
            yloc = float(att.get('yloc', '') )
            node = self.top_pile.add_simple_node(name, xloc, yloc)
            self.node_dict[name] = node

        elif name == "CPermNode":
            name = att.get('name', '').encode('ascii')
            xloc = float(att.get('xloc', '') )
            yloc = float(att.get('yloc', '') )
            node = self.top_pile.add_perm_node(name, xloc, yloc)
            self.node_dict[name] = node

        elif name == "CInputNode":
            name = att.get('name', '').encode('ascii')
            xloc = float(att.get('xloc', '') )
            yloc = float(att.get('yloc', '') )
            node = self.top_pile.add_input_node(name, xloc, yloc)
            self.node_dict[name] = node


        elif name == "CMacroNode":
            name = att.get('name', '').encode('ascii')
            xloc = float(att.get('xloc', '') )
            yloc = float(att.get('yloc', '') )
            wloc = float(att.get('wloc', '') )
            hloc = float(att.get('hloc', '') )

            node = self.top_pile.add_macro_subnode(name, xloc, yloc,
                                                   wloc, hloc)
            self.node_dict[name] = node

            self.pile_node.append(node)
            # symbol table put on stack to preserve macro scope for inputs
            new_node_dict = dict()
            self.pile_dict.append(new_node_dict)
            self.top_pile = node
            self.node_dict = new_node_dict

        elif name == "transition":
            name = att.get('name', '').encode('ascii')
            ori = att.get('ori', '')
            ext = att.get('ext', '')
            event = att.get('event', '')
            condition = att.get('condition', '')
            action = att.get('action', '')
            fact_ids_text = att.get('fact_ids','')[1:-1]
            if len(fact_ids_text) > 0:
                fact_ids_split = fact_ids_text.split(',')
                fact_ids = []
                for fid in fact_ids_split:
                    fact_ids.append(int(fid))
            else:
                fact_ids = []

            try:
                node_ori = self.node_dict[ori]
                node_ext = self.node_dict[ext]
            except Exception, exc:
                print 'Bad xml file - missing nodes', ori, ' ', ext
                print self.node_dict
                print exc

            self.transition = self.top_pile.add_transition(node_ori, node_ext)
            # the transition may not be created (origin = ext for example)
            if self.transition:
                self.transition.set_event(event)
                self.transition.set_condition(condition)
                self.transition.set_action(action)
                self.transition.fact_ids = fact_ids

                self.in_transition = True
                self.transition.note = ""

        elif name == 'constraints':
            self.in_constraints = True
            self.constraints = ""

        else: # ignore
            pass

    def characters(self, chr):
        """
        xxx
        """
        if self.in_constraints:
            self.constraints = self.constraints + chr
        if self.in_transition:
            self.transition.note = self.transition.note + chr

    def endElement (self, name):
        """
        xxx
        """
        if name == "model":
            return

        elif name == "CMacroNode":
            #self.top_pile = self.pile_node.pop()
            self.pile_node.remove(self.top_pile)
            self.top_pile = self.pile_node[-1]
            #self.node_dict = self.pile_dict.pop()
            self.pile_dict.remove(self.node_dict)
            self.node_dict = self.pile_dict[-1]

        elif name == 'constraints':
            self.in_constraints = False
            self.model.constraints = self.constraints + '\n'

        elif name == 'transition':
            self.in_transition = False
#            if self.transition:
#                self.transition.note = self.transition.note + '\n'
        else:
            pass


class MakeModelFromXmlFile:
    """
    parse a xml file
    """
    def __init__(self, xml_file, model = None):
        self.model = model
        self.handler = MakeHandler(model=self.model)
        self.parser = make_parser()
        self.parser.setContentHandler(self.handler)

        try:
            self.parser.parse(xml_file)
        except Exception, exc:
            print 'ERROR while xml parsing'
            print exc


    def get_model(self):
        """
        As it says
        """
        return self.handler.model

class MakeModelFromXmlString:
    """
    parse a xml description as string
    """
    def __init__(self, xml_string):
        self.model = None
        self.handler = MakeHandler()
        self.parser = make_parser()
        self.parser.setContentHandler(self.handler)

        try:
            PS(xml_string, self.handler)
        except Exception, exc:
            print 'ERROR while xml parsing'
            print exc


    def get_model(self):
        """
        As it says
        """
        return self.handler.model


class XmlException(Exception):
    """
    For exception identification
    """
    def __init__(self, mess):
        self.message = mess

























##
## Filename    : input_stream.py
## Author(s)   : Michel Le Borgne
## Created     : 11/2009
## Revision    :
## Source      :
##
## Copyright 2009 : IRISA/IRSET
##
## This library is free software; you can redistribute it and/or modify it
## under the terms of the GNU General Public License as published
## by the Free Software Foundation; either version 2.1 of the License, or
## any later version.
##
## This library is distributed in the hope that it will be useful, but
## WITHOUT ANY WARRANTY, WITHOUT EVEN THE IMPLIED WARRANTY OF
## MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  The software and
## documentation provided hereunder is on an "as is" basis, and IRISA has
## no obligations to provide maintenance, support, updates, enhancements
## or modifications.
## In no event shall INRIA be liable to any party for direct, indirect,
## special, incidental or consequential damages, including lost profits,
## arising out of the use of this software and its documentation, even if
## IRISA have been advised of the possibility of such damage.  See
## the GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this library; if not, write to the Free Software Foundation,
## Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
##
## The original code contained here was initially developed by:
##
##     Michel Le Borgne.
##     IRISA
##     Symbiose team
##     IRISA  Campus de Beaulieu
##     35042 RENNES Cedex, FRANCE
##
##
## Contributor(s): Geoffroy Andrieux, Nolwenn Le Meur
##

"""
Stream to manage input/clocks during a simulation
"""

import string
from string import  atoi
from cadbiom.models.guard_transitions.simulator.simul_exceptions import \
                               SimulException, SimulStopException

class InputStream(object):
    """
    General class for all input buffers. Input buffers manage dynamic discrete systems inputs
    It has essentially two methods: next_input which deliver an input as a list of activated inputs
    and rewind which initializes the buffer
    """
    def __init__(self):
        self.current_index = 0    # line pointer in buffer
        self.max_index = 0        # size of input buffer

    def next_input(self):
        """
        returns a list of activated inputs (others are inactivated)
        """
        return []

    def rewind(self):
        """
        rewind the input stream for a new simulation
        """
        self.current_index = 0

    def is_all_read(self):
        """
        test end of buffer
        """
        return self.current_index == self.max_index



class CodedStream(InputStream):
    """
    InputStream where inputs are described with a language
    """
    def __init__(self):
        self.current_index = 0    # line pointer in buffer
        self.max_index = 0        # size of input buffer
        self.input_buffer = []    # main buffer
        self.nb_inputs = 0        # number of inputs per line
        self.input_names = []     # list of input names ordered as in first line
        self.label_table = dict() # label name -> address
        self.block_count = None


    def next_input(self):
        """
        Produce the next input as a list of activated inputs (including clocks)
        """
        # end of buffer
        if self.current_index >= self.max_index:
            self.__load_input_buffer()

        line_split = self.input_buffer[self.current_index].split()

        # line analysis: command
        while line_split[0][0] == '$':
            self.__execute_command(line_split) # may change current index
            line_split = self.input_buffer[self.current_index].split()

        # ordinary input
        if len(line_split) != self.nb_inputs:
            mess = "Number of inputs different from nb of declared inputs"
            raise SimulException(mess)
        out = []
        for i in range(self.nb_inputs):
            val = int(line_split[i])
            if val == 1:
                out.append(self.input_names[i])
        self.current_index = self.current_index + 1
        return out

    def __execute_command(self, line_split):
        """
        implements command execution for coded steam
        """
        command = line_split[0][1:]
        if command == 'label':
            self.label_table[line_split[1]] = (self.current_index, 0)
            self.current_index = self.current_index + 1
            return

        elif command == 'repeat':
            label = line_split[1]
            if not self.label_table.has_key(label): # label doesn't exist
                raise SimulException("Unknown label line %s"%self.current_index)
            (index, count) = self.label_table[label]
            time = atoi(line_split[2]) # number of loops
            if count < time:
                self.current_index = index+1
                self.label_table[label] = (index, count+1)
            else:
                self.current_index = self.current_index + 1
                self.label_table[label] = (index, 0)
            return

        elif command == 'block':
            if self.current_index == 1:
                mess = "Input Syntax error line %$: Block on first line,"
                mess = mess + " value must be before" % self.current_index
                raise SimulException(mess)
            if self.block_count == None:
                self.block_count = atoi(line_split[1])
            if self.block_count == 0: # end of block
                self.block_count = None
                self.current_index = self.current_index + 1
            elif self.block_count == -1: # continuous block
                self.current_index = self.current_index - 1
            else: # limited block continues
                self.current_index = self.current_index - 1
                self.block_count = self.block_count - 1

        elif command == 'stop':
            raise SimulStopException()

        else:
            raise SimulException("Unknown command line %s"%self.current_index)

    def __load_input_buffer(self):
        """
        not implemented - for partial loading ??
        """
        raise SimulStopException()

    def rewind(self):
        """
        as it says
        """
        self.current_index = 1
        self.block_count = None

    def is_all_read(self):
        """
        test end of buffer
        """
        return self.current_index == self.max_index



class CFileInput(CodedStream):
    """
    Input buffer with input description in file
    """
    def __init__(self, filename):
        self.file = open(filename, 'r')
        self.input_buffer = self.file.readlines() # read all the lines
        self.max_index = len(self.input_buffer)
        self.current_index = 1               # line 0 is for input description
        line = string.strip(self.input_buffer[0]) # get line 0
        line_split = string.split(line)
        if line_split[0] != "$inputs":
            raise SimulException()
        else:
            self.input_names = []            # collect input names
            for name in line_split[1:]:
                self.input_names.append(name)
            self.nb_inputs = len(self.input_names)
        self.label_table = dict()
        self.block_count = None



class ActInput(InputStream):
    """
    Input buffer where input names are enumerated in a file (pure input or scenario)
    or
    """
    def __init__(self, filename = None, inlist = []):
        if filename:
            ifile = open(filename, 'r')
            self.input_buffer = ifile.readlines() # read all the lines
            ifile.close()
        else:
            self.input_buffer = inlist
        self.max_index = len(self.input_buffer)
        if self.input_buffer[0][0] == '%': # pure input file
            self.current_index = 0
            self.init_index = 0
        else: # scenario
            self.current_index = 1
            self.init_index = 1

    def next_input(self):
        """
        Produce the next input as a list of activated inputs (including clocks)
        coded in file as "% h1 i2 h3"
        """
        # end of buffer
        if self.current_index >= self.max_index:
            raise SimulStopException()
        cin =  self.input_buffer[self.current_index]
        self.current_index += 1
        return cin.split()[1:]

    def rewind(self):
        self.current_index = self.init_index




## Filename    : MCLQuery.py
## Author(s)   : Michel Le Borgne
## Created     : 10 sept. 2012
## Revision    :
## Source      :
##
## Copyright 2012 : IRISA/IRSET
##
## This library is free software; you can redistribute it and/or modify it
## under the terms of the GNU General Public License as published
## by the Free Software Foundation; either version 2.1 of the License, or
## any later version.
##
## This library is distributed in the hope that it will be useful, but
## WITHOUT ANY WARRANTY, WITHOUT EVEN THE IMPLIED WARRANTY OF
## MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  The software and
## documentation provided here under is on an "as is" basis, and IRISA has
## no obligations to provide maintenance, support, updates, enhancements
## or modifications.
## In no event shall IRISA be liable to any party for direct, indirect,
## special, incidental or consequential damages, including lost profits,
## arising out of the use of this software and its documentation, even if
## IRISA have been advised of the possibility of such damage.  See
## the GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this library; if not, write to the Free Software Foundation,
## Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
##
## The original code contained here was initially developed by:
##
##     Michel Le Borgne.
##     IRISA
##     Symbiose team
##     IRISA  Campus de Beaulieu
##     35042 RENNES Cedex, FRANCE
##
##
## Contributor(s): Geoffroy Andrieux - IRISA/IRSET
##
"""
Query internal representation
"""

from cadbiom.models.clause_constraints.mcl.MCLSolutions import MCLException
from cadbiom import commons as cm
from logging import DEBUG

LOGGER = cm.logger()

class MCLSimpleQuery(object):
    """
    Class packaging the elements of a query:
        - start property (string)
        - invariant property (string)
        - final property (sting)
        - dim_start: a start property in DIMACS form - optional - default None
        - dim_invariant: idem for invariant property
        - dim_final: idem for final property
        - steps_before _reach: number of shift before testing the final property
                               optional - default value 1000 000
    NB: DClause: a clause coded as a list of DIMACS coded literals
    """

    def __init__(self, start_prop, inv_prop, final_prop):
        """
        @param start_prop: init property - None is allowed
        @param inv_prop: invariant property - None allowed
        @param final_prop: final property - None is allowed
        """
        if LOGGER.getEffectiveLevel() == DEBUG:
            LOGGER.debug("MCLSimpleQuery params:: " + \
                         str(start_prop) + '; ' + \
                         str(inv_prop) + '; ' + \
                         str(final_prop))
        self.start_prop = start_prop  # logical formula or None
        self.inv_prop = inv_prop      # logical formula or None
        self.variant_prop = []        # list<logical formula>
        self.final_prop = final_prop  # logical formula or None
        # dimacs form of preceding
        self.dim_start = []         # list<DClause>
        self.dim_inv = []           # list<DClause>
        self.dim_final = []         # list<DClause>
        self.dim_variant_prop = []  # list<list<DClause>>

        self.steps_before_reach = 0

    @classmethod
    def from_frontier_sol(cls, f_sol):
        """
        build a query from a frontier solution
        start condition activate the activated frontier places
        variant property enforce same timing
        """
        # start condition enforce activation of solution places
        st_prop = None
        if f_sol.activated_frontier:
            st_prop = f_sol.activated_frontier[0]
            for fcond in f_sol.activated_frontier[1:]:
                st_prop = st_prop + " and " + fcond
        # no invariant property
        inv_prop = None
        # variant property enforce same timing
        var_prop = None
        if f_sol.ic_sequence:
            var_prop = []
            for ic_act in f_sol.ic_sequence:
                l_ic = ic_act[1:].split()
                if len(l_ic)>0:
                    ic_prop = l_ic[0]
                    for icp in l_ic[1:]:
                        ic_prop = ic_prop + " and " + icp
                else:
                    ic_prop = ""
                var_prop.append(ic_prop)
        fin_prop = None
        n_query = MCLSimpleQuery(st_prop, inv_prop, fin_prop)
        n_query.set_variant_property(var_prop)
        return n_query


    @classmethod
    def from_frontier_sol_new_timing(cls, f_sol, unfolder):
        """
        @param f_sol: the FrontierSolution
        @param unfolder: unfolder in use
        build a query from a frontier solution
        start condition activate the activated frontier places
        and force others to be inactivate.
        variant property enforce new timing
        """
        # activation part of start condition
        st_prop = None
        if f_sol.activated_frontier:
            st_prop = f_sol.activated_frontier[0]
            for fcond in f_sol.activated_frontier[1:]:
                st_prop = st_prop + " and " + fcond
        # no invariant property
        inv_prop = None
        # variant property enforce new timing
        var_prop = None
        if f_sol.ic_sequence:
            var_prop = []
            for ic_act in f_sol.ic_sequence:
                l_ic = ic_act[1:].split()
                if len(l_ic)>0:
                    ic_prop = "not (" + l_ic[0]+ ")" # step constraint
                    for icp in l_ic[1:]:
                        n_icp = "not (" + icp + ")"
                        ic_prop = ic_prop + " or " + n_icp
                else:
                    ic_prop = ""
                var_prop.append(ic_prop)
        fin_prop = None
        # inactivation of other frontier places at start (dimacs form)
        frontier = unfolder.get_frontier()
        l_dstart = []
        for i_fr in frontier:
            name = unfolder.get_var_name(i_fr)
            if not name in f_sol.activated_frontier:
                l_dstart.append([- i_fr])

        n_query = MCLSimpleQuery(st_prop, inv_prop, fin_prop)
        n_query.set_variant_property(var_prop)
        n_query.set_dim_start(l_dstart)
        return n_query

    @classmethod
    def from_frontier_sol_same_timing(cls, f_sol, unfolder):
        """
        @param f_sol: the FrontierSolution
        @param unfolder: unfolder in use
        build a query from a frontier solution
        start condition activate the activated frontier places
        and force others to be inactivate.
        variant property enforce same timing on activated events,
        and others are free
        """
        # activation part of start condition
        st_prop = None
        if f_sol.activated_frontier:
            st_prop = f_sol.activated_frontier[0]
            for fcond in f_sol.activated_frontier[1:]:
                st_prop = st_prop + " and " + fcond
        # no invariant property
        inv_prop = None
        # make event dictionary
        event_dict = dict()
        step_cpt = 0
        for ic_act in f_sol.ic_sequence:
            l_ic = ic_act[1:].split()
            for icp in l_ic:
                if not event_dict.has_key(icp):
                    event_dict[icp]=[step_cpt]
                else :
                    event_dict[icp].append(step_cpt)
            step_cpt += 1
        # variant property enforce same timing on activated events
        var_prop = None
        if f_sol.ic_sequence:
            var_prop = []
            step_cpt = 0
            for ic_act in f_sol.ic_sequence:
                l_ic = ic_act[1:].split()
                if len(l_ic)>0:
                    if event_dict.has_key(l_ic[0]):
                        if step_cpt in event_dict[l_ic[0]]:
                            ic_prop = l_ic[0]
                        else :
                            ic_prop = "not (" + l_ic[0]+ ")"
                    for icp in l_ic[1:]:
                        if event_dict.has_key(icp):
                            if step_cpt in event_dict[icp]:
                                n_icp = icp
                            else :
                                n_icp = "not (" + icp + ")"
                        ic_prop = ic_prop + " and " + n_icp
                else:
                    ic_prop = ""
                var_prop.append(ic_prop)
                step_cpt += 1
        fin_prop = None
        # inactivation of other frontier places at start (dimacs form)
        frontier = unfolder.get_frontier()
        l_dstart = []
        for i_fr in frontier:
            name = unfolder.get_var_name(i_fr)
            if not name in f_sol.activated_frontier:
                l_dstart.append([- i_fr])

        n_query = MCLSimpleQuery(st_prop, inv_prop, fin_prop)
        n_query.set_variant_property(var_prop)
        n_query.set_dim_start(l_dstart)
        return n_query

    def set_variant_property(self, var_prop):
        """
        @param var_prop: list<string> - list of logical formulas
        """
        self.variant_prop = var_prop

    def set_dim_variant_prop(self, d_var_prop):
        """
        @param d_var_prop: list<list<list<int>>> - list-list of dimacs clauses
        """
        self.dim_variant_prop = d_var_prop

    def set_dim_start(self, dim_start):
        """
        @param dim_start: start property in DIMACS format
        """
        self.dim_start = dim_start

    def set_dim_inv(self, dim_inv):
        """
        Similar to set_dim_start
        """
        self.dim_inv = dim_inv

    def set_dim_final(self, dim_final):
        """
        Similar to set_dim_final
        """
        self.dim_final = dim_final

    def set_steps_before_reach(self, nb_steps):
        """
        @param nb_steps: int - number of shifts before testing final prop
        """
        self.steps_before_reach = nb_steps

    def merge(self, quer):
        """
        Merge two queries into one. start, inv and final properties are
        the and of the corresponding properties in each query.
        If both queries have variant properties, they must be on the same horizon.
        steps before reach is set to zero.

        @param quer: a MCLSimpleQuery
        @raise MCLException: if the queries have variant properties on
        different horizons.
        """
        # merge of start properties
        if not self.start_prop:
            st_prop = quer.start_prop
        elif not quer.start_prop:
            st_prop = self.start_prop
        else:
            st_prop = "(" + self.start_prop + ") and (" + quer.start_prop + ")"
        # merge of invariant properties
        if not self.inv_prop:
            i_prop = quer.inv_prop
        elif not quer.inv_prop:
            i_prop = self.inv_prop
        else:
            i_prop = "(" + self.inv_prop + ") and (" + quer.inv_prop + ")"
        # merge of final properties
        if not self.final_prop:
            f_prop = quer.final_prop
        elif not quer.final_prop:
            f_prop = self.final_prop
        else:
            f_prop = "(" + self.final_prop + ") and (" + quer.final_prop + ")"

        n_query = MCLSimpleQuery(st_prop, i_prop, f_prop)
        # merge of variant prop
        v_prop = None
        if not self.variant_prop:
            v_prop = quer.variant_prop
        elif not quer.variant_prop:
            v_prop = self.variant_prop
        else:
            ll1 = len(self.variant_prop)
            ll2 = len(quer.variant_prop)
            if ll1 != ll2:
                mess = "Tempting to merge two queries with variant prop "
                mess = mess + "of different length"
                raise MCLException(mess)
            else:
                v_prop = []
                for i in range(ll1):
                    loc = self.variant_prop[i] + " and "+ quer.variant_prop[i]
                    v_prop.append(loc)
        n_query.set_variant_property(v_prop)
        # merge of dim properties
        dim_start = self.dim_start + quer.dim_start
        n_query.set_dim_start(dim_start)
        dim_inv = self.dim_inv + quer.dim_inv
        n_query.set_dim_inv(dim_inv)
        dim_final = self.dim_final + quer.dim_final
        n_query.set_dim_final(dim_final)

        return n_query

    def __str__(self):
        """
        Print logical formulas in query
        """
        str_out = "Start_property: "
        if self.start_prop:
            str_out = str_out + self.start_prop
        str_out = str_out + "\nInv property: "
        if self.inv_prop:
            str_out = str_out + self.inv_prop
        str_out = str_out + "\nFinal property: "
        if self.final_prop:
            str_out = str_out + self.final_prop
        str_out = str_out + "\nVariant property: "
        if self.variant_prop:
#            str_out = str_out + self.variant_prop[0]
#            for prop in self.variant_prop[1:]:
#                str_out = str_out + ", " + prop
            str_out = str_out + self.variant_prop.__str__()
        return str_out



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (C) 2010-2017  IRISA
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
#
# The original code contained here was initially developed by:
#
#     Geoffroy Andrieux.
#     IRISA/IRSET
#     Symbiose team
#     IRISA  Campus de Beaulieu
#     35042 RENNES Cedex, FRANCE
#
#
# Contributor(s): Michel Le Borgne, Nolwenn Le Meur, Pierre Vignet

"""Definition of setup function for setuptools module."""

# Standard imports
import sys
from distutils import sysconfig
from setuptools import setup, Extension, find_packages
from setuptools.command.test import test as TestCommand

__PACKAGE_VERSION__ = "0.1.2"
__LIBRARY_VERSION__ = "0.1.0"

################################################################################

class PyTest(TestCommand):
    """Call tests with the custom 'python setup.py test' command."""

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.pytest_args = []

    def run_tests(self):
        import pytest
        errno = pytest.main()
        sys.exit(errno)

################################################################################

# Delete unwanted flags for C compilation
# Distutils has the lovely feature of providing all the same flags that
# Python was compiled with. The result is that adding extra flags is easy,
# but removing them is a total pain. Doing so involves subclassing the
# compiler class, catching the arguments and manually removing the offending
# flag from the argument list used by the compile function.
# That's the theory anyway, the docs are too poor to actually guide you
# through what you have to do to make that happen.
d = sysconfig.get_config_vars()
for k, v in d.items():
    for unwanted in ('-Wstrict-prototypes', '-DNDEBUG', ' -g ',
                     '-O2', '-D_FORTIFY_SOURCE=2', '-fstack-protector-strong'):
        if str(v).find(unwanted) != -1:
            v = d[k] = str(v).replace(unwanted, ' ')

modules = [
    Extension(
        "_cadbiom",
        ["_cadbiom/cadbiom.c"],
        language = "c",
        include_dirs=['_cadbiom', '.'],
        define_macros=[
            ('CADBIOM_VERSION', '"' + __LIBRARY_VERSION__ + '"'),
        ],
        extra_compile_args=[
            "-flto",
            "-march=native",
            "-mtune=native",
            "-Ofast",
            #"-O3",
            #"-Wall",
            #"-g", # Not define NDEBUG macro => Debug build
            "-DNDEBUG", # Force release build
            #"-DBOOST_TEST_DYN_LINK",
            #"-DUSE_ZLIB",
            "-std=c11",
            "-Wno-unused-variable",
            "-Wno-unused-but-set-variable",
            # assume that signed arithmetic overflow of addition, subtraction
            # and multiplication wraps around using twos-complement
            # representation
            "-fwrapv",
            #BOF protect (use both)
            #"-D_FORTIFY_SOURCE=2",
            #"-fstack-protector-strong",
            "-pthread",
            "-DUSE_PTHREADS",
            "-fopenmp",
            "-D_GLIBCXX_PARALLEL",
        ],
        extra_link_args=[
            "-Ofast",
            "-march=native",
            "-flto",
            "-std=c11",
            "-fopenmp",
        ]
    ),
]

setup(
    name="cadbiom",
    version=__PACKAGE_VERSION__,
    author="Pierre Vignet",
    author_email="pierre.vignet@irisa.fr",
    url = "https://gitlab.irisa.fr/0000B8EG/Cadbiom/tree/pypi_packaging",

    # Search all packages recursively
    packages=find_packages(),

    classifiers=[
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: GNU General Public License v2 or later (GPLv2+)",
        "Natural Language :: French",
        "Intended Audience :: Developers",
        "Operating System :: OS Independent",
        "Programming Language :: C",
        "Programming Language :: Python :: 2.7",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
    ],
    ext_modules=modules,
    description="Cadbiom library",
    long_description=open('README.md').read(),
    install_requires=["networkx<2", "lxml", "pycryptosat"],

    # Tests
    tests_require=['pytest'],
    cmdclass={'test': PyTest},
)
